# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Callable, Iterable, Sequence
import typing
from typing import Any, TypeAlias, Union, get_args

from typing_extensions import Self

from ansys.api.acp.v0 import sublaminate_pb2, sublaminate_pb2_grpc

from .._utils.property_protocols import ReadOnlyProperty, ReadWriteProperty
from ._grpc_helpers.edge_property_list import (
    GenericEdgePropertyType,
    define_add_method,
    define_edge_property_list,
)
from ._grpc_helpers.polymorphic_from_pb import tree_object_from_resource_path
from ._grpc_helpers.property_helper import (
    _exposed_grpc_property,
    grpc_data_property,
    grpc_data_property_read_only,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .enums import SymmetryType, status_type_from_pb, symmetry_type_from_pb, symmetry_type_to_pb
from .fabric import Fabric
from .object_registry import register
from .stackup import Stackup

__all__ = ["SubLaminate", "Lamina"]

_LINKABLE_MATERIAL_TYPES: TypeAlias = Union[Fabric, Stackup]


@mark_grpc_properties
class Lamina(GenericEdgePropertyType):
    """
    Class to link a material with a sub-laminate.

    Parameters
    ----------
    material :
        Link to an existing fabric or stackup.
    angle :
        Orientation angle in degree of the material with respect to the reference direction.

    """

    _SUPPORTED_SINCE = "24.2"

    def __init__(self, material: _LINKABLE_MATERIAL_TYPES, angle: float = 0.0):
        self._callback_apply_changes: Callable[[], None] | None = None
        self.material = material
        self.angle = angle

    @_exposed_grpc_property
    def material(self) -> _LINKABLE_MATERIAL_TYPES:
        """Link to an existing fabric or stackup."""
        return self._material

    @material.setter
    def material(self, value: _LINKABLE_MATERIAL_TYPES) -> None:
        if not isinstance(value, get_args(_LINKABLE_MATERIAL_TYPES)):
            raise TypeError(
                f"Expected material to be of type {get_args(_LINKABLE_MATERIAL_TYPES)}, "
                f"got {type(value)}."
            )
        self._material: _LINKABLE_MATERIAL_TYPES = value
        if self._callback_apply_changes:
            self._callback_apply_changes()

    @_exposed_grpc_property
    def angle(self) -> float:
        """Orientation angle in degree of the material with respect to the reference direction."""
        return self._angle

    @angle.setter
    def angle(self, value: float) -> None:
        self._angle = value
        if self._callback_apply_changes:
            self._callback_apply_changes()

    def _set_callback_apply_changes(self, callback_apply_changes: Callable[[], None]) -> None:
        self._callback_apply_changes = callback_apply_changes

    @classmethod
    def _from_pb_object(
        cls,
        parent_object: CreatableTreeObject,
        message: sublaminate_pb2.Lamina,
        apply_changes: Callable[[], None],
    ) -> Lamina:
        material = tree_object_from_resource_path(
            resource_path=message.material, server_wrapper=parent_object._server_wrapper
        )

        if not isinstance(material, get_args(_LINKABLE_MATERIAL_TYPES)):
            raise TypeError(
                f"Expected material to be of type {get_args(_LINKABLE_MATERIAL_TYPES)}, "
                f"got {type(material)}."
            )

        material = typing.cast("_LINKABLE_MATERIAL_TYPES", material)
        new_obj = cls(
            material=material,
            angle=message.angle,
        )
        new_obj._set_callback_apply_changes(apply_changes)
        return new_obj

    def _to_pb_object(self) -> sublaminate_pb2.Lamina:
        return sublaminate_pb2.Lamina(material=self.material._resource_path, angle=self.angle)

    def _check(self) -> bool:
        # Check for empty resource paths
        return bool(self.material._resource_path.value)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            return (
                self.material._resource_path == other.material._resource_path
                and self.angle == other.angle
            )

        return False

    def __repr__(self) -> str:
        return f"Lamina(material={self.material.__repr__()}, angle={self.angle})"

    def clone(self) -> Self:
        """Create a new unstored Lamina with the same properties."""
        return type(self)(material=self.material, angle=self.angle)


@mark_grpc_properties
@register
class SubLaminate(CreatableTreeObject, IdTreeObject):
    """Instantiate a SubLaminate.

    Parameters
    ----------
    name :
        Name of the sub-laminate.
    symmetry :
        Whether the sub-laminate is odd or even symmetrical, or none.
    topdown :
        The first material in the list is placed first in the mold if topdown is true.
    materials :
        List of materials (fabrics and stackups) with angles which build the sub-laminate.
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "sublaminates"
    _OBJECT_INFO_TYPE = sublaminate_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = sublaminate_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "SubLaminate",
        symmetry: SymmetryType = "no_symmetry",
        topdown: bool = True,
        materials: Sequence[Lamina] = tuple(),
    ):
        super().__init__(name=name)

        self.symmetry = SymmetryType(symmetry)
        self.topdown = topdown
        self.materials = materials

    def _create_stub(self) -> sublaminate_pb2_grpc.ObjectServiceStub:
        return sublaminate_pb2_grpc.ObjectServiceStub(self._channel)

    locked: ReadOnlyProperty[bool] = grpc_data_property_read_only("properties.locked")
    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    thickness: ReadOnlyProperty[float] = grpc_data_property_read_only("properties.thickness")
    area_weight: ReadOnlyProperty[float] = grpc_data_property_read_only("properties.area_weight")
    area_price: ReadOnlyProperty[float] = grpc_data_property_read_only("properties.area_price")

    symmetry = grpc_data_property(
        "properties.symmetry",
        from_protobuf=symmetry_type_from_pb,
        to_protobuf=symmetry_type_to_pb,
    )
    topdown: ReadWriteProperty[bool, bool] = grpc_data_property("properties.topdown")

    materials = define_edge_property_list("properties.materials", Lamina)
    add_material = define_add_method(
        Lamina,
        attribute_name="materials",
        func_name="add_material",
        parent_class_name="SubLaminate",
        module_name=__module__,
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable, Sequence
import typing

from ansys.api.acp.v0 import field_definition_pb2, field_definition_pb2_grpc

from .._utils.property_protocols import ReadWriteProperty
from ._grpc_helpers.linked_object_list import define_polymorphic_linked_object_list
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .element_set import ElementSet
from .enums import status_type_from_pb
from .lookup_table_1d_column import LookUpTable1DColumn
from .lookup_table_3d_column import LookUpTable3DColumn
from .modeling_ply import ModelingPly
from .object_registry import register
from .oriented_selection_set import OrientedSelectionSet

__all__ = ["FieldDefinition"]


_SCOPE_ENTITIES_LINKABLE_TO_FIELD_DEFINITION: typing.TypeAlias = typing.Union[
    ElementSet,
    OrientedSelectionSet,
    ModelingPly,
]


@mark_grpc_properties
@register
class FieldDefinition(CreatableTreeObject, IdTreeObject):
    """Instantiate a Field Definition.

    A field definition is used to configure the state of variable materials.
    For instance, a Lookup Table can be used to define the distribution of a
    state of material field (e.g. degradation). The field definition allows
    to define the material field per element or per ply and element.

    Note: Field definitions are currently only supported through (Py)Mechanical.
    The direct interface of PyACP to (Py)MADL ignores field definitions.

    Parameters
    ----------
    name :
        Name of the field definition.
    active :
        Inactive field definitions are ignored.
    field_variable_name :
        Links the material field to a field variable name.
        The field variable name must be defined in the material properties.
        Note that the ``Temperature`` and ``Shear Angle`` field variables are not available
        for Field Definitions. Temperature is defined through the solution and `Shear Angle`
        is defined via draping calculations.
    scope_entities :
        Define the scope of the field definition. You can use a combination
        of Element Sets and Oriented Selection Sets for elemental scope.
        For ply-wise field definitions, a combination of Modeling Plies can be selected.
        Note that the field definition is only applied to the Analysis Plies attached to
        the Modeling Plies. A combination of elemental and ply-wise definition is not supported.
    scalar_field :
        Select the scalar Look-Up Table column from which the state of the field
        variable is interpolated from.
    full_mapping :
        Specify to include the shell offset of each analysis ply for the interpolation process.
        The default is to interpolate the state of the field variables at the shell element centroid.
        For solid elements, the position of each analysis ply is automatically considered.

    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "field_definitions"
    _OBJECT_INFO_TYPE = field_definition_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = field_definition_pb2.CreateRequest
    _SUPPORTED_SINCE = "25.1"

    def __init__(
        self,
        *,
        name: str = "FieldDefinition",
        active: bool = True,
        field_variable_name: str = "",
        scope_entities: Sequence[_SCOPE_ENTITIES_LINKABLE_TO_FIELD_DEFINITION] = tuple(),
        scalar_field: LookUpTable1DColumn | LookUpTable3DColumn | None = None,
        full_mapping: bool = False,
    ):
        super().__init__(name=name)

        self.active = active
        self.field_variable_name = field_variable_name
        self.scope_entities = scope_entities
        self.scalar_field = scalar_field
        self.full_mapping = full_mapping

    def _create_stub(self) -> field_definition_pb2_grpc.ObjectServiceStub:
        return field_definition_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    active: ReadWriteProperty[bool, bool] = grpc_data_property("properties.active")
    field_variable_name: ReadWriteProperty[str, str] = grpc_data_property(
        "properties.field_variable_name"
    )

    scope_entities = define_polymorphic_linked_object_list(
        "properties.scope_entities",
        allowed_types=(
            ElementSet,
            OrientedSelectionSet,
            ModelingPly,
        ),
    )

    scalar_field = grpc_link_property(
        "properties.scalar_field", allowed_types=(LookUpTable1DColumn, LookUpTable3DColumn)
    )

    full_mapping: ReadWriteProperty[bool, bool] = grpc_data_property("properties.full_mapping")



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable

from ansys.api.acp.v0 import cut_off_geometry_pb2, cut_off_geometry_pb2_grpc

from .._utils.property_protocols import ReadWriteProperty
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .enums import (
    CutOffGeometryOrientationType,
    cut_off_geometry_orientation_type_from_pb,
    cut_off_geometry_orientation_type_to_pb,
    status_type_from_pb,
)
from .object_registry import register
from .virtual_geometry import VirtualGeometry

__all__ = ["CutOffGeometry"]


@mark_grpc_properties
@register
class CutOffGeometry(CreatableTreeObject, IdTreeObject):
    """Instantiate a cut-off geometry.

    Parameters
    ----------
    name :
        Name of the cut-off geometry.
    active :
        Inactive cut-off geometries are not used in the solid model extrusion.
    cad_geometry :
        The geometry defining the cut-off.
    orientation_type :
        Determines the cutting orientation of a surface/body geometry. Allows to
        switch between include and exclude.
    relative_merge_tolerance :
        Set the merging tolerance for neighboring nodes relative to the element size.
    """

    __slots__: Iterable[str] = ()

    _COLLECTION_LABEL = "cut_off_geometries"
    _OBJECT_INFO_TYPE = cut_off_geometry_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = cut_off_geometry_pb2.CreateRequest
    _SUPPORTED_SINCE = "25.1"

    def __init__(
        self,
        *,
        name: str = "CutOffGeometry",
        active: bool = True,
        cad_geometry: VirtualGeometry | None = None,
        orientation_type: CutOffGeometryOrientationType = CutOffGeometryOrientationType.UP,
        relative_merge_tolerance: float = 0.1,
    ):
        super().__init__(
            name=name,
        )
        self.active = active
        self.cad_geometry = cad_geometry
        self.orientation_type = orientation_type
        self.relative_merge_tolerance = relative_merge_tolerance

    def _create_stub(self) -> cut_off_geometry_pb2_grpc.ObjectServiceStub:
        return cut_off_geometry_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    active: ReadWriteProperty[bool, bool] = grpc_data_property("properties.active")
    cad_geometry = grpc_link_property("properties.cad_geometry", allowed_types=(VirtualGeometry,))
    orientation_type = grpc_data_property(
        "properties.orientation",
        from_protobuf=cut_off_geometry_orientation_type_from_pb,
        to_protobuf=cut_off_geometry_orientation_type_to_pb,
    )
    relative_merge_tolerance: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.relative_merge_tolerance"
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable, Sequence

from ansys.api.acp.v0 import (
    analysis_ply_pb2_grpc,
    enum_types_pb2,
    imported_analysis_ply_pb2_grpc,
    layup_mapping_object_pb2,
    layup_mapping_object_pb2_grpc,
)

from .._utils.property_protocols import ReadWriteProperty
from ._grpc_helpers.enum_wrapper import wrap_to_string_enum
from ._grpc_helpers.linked_object_list import (
    define_linked_object_list,
    define_polymorphic_linked_object_list,
)
from ._grpc_helpers.mapping import get_read_only_collection_property
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from .analysis_ply import AnalysisPly
from .base import CreatableTreeObject, IdTreeObject
from .element_set import ElementSet
from .enums import (
    BaseElementMaterialHandling,
    ElementTechnology,
    ReinforcingBehavior,
    StressStateType,
    base_element_material_handling_from_pb,
    base_element_material_handling_to_pb,
    element_technology_from_pb,
    element_technology_to_pb,
    reinforcing_behavior_from_pb,
    reinforcing_behavior_to_pb,
    status_type_from_pb,
    stress_state_type_from_pb,
    stress_state_type_to_pb,
)
from .imported_analysis_ply import ImportedAnalysisPly
from .imported_modeling_group import ImportedModelingGroup
from .imported_modeling_ply import ImportedModelingPly
from .material import Material
from .modeling_group import ModelingGroup
from .modeling_ply import ModelingPly
from .object_registry import register
from .rosette import Rosette
from .solid_element_set import SolidElementSet

__all__ = ["LayupMappingObject", "LayupMappingRosetteSelectionMethod"]


(
    LayupMappingRosetteSelectionMethod,
    layup_mapping_rosette_selection_method_to_pb,
    layup_mapping_rosette_selection_method_from_pb,
) = wrap_to_string_enum(
    "LayupMappingRosetteSelectionMethod",
    enum_types_pb2.RosetteSelectionMethod,
    module=__name__,
    doc="Options for how the rosette is selected in the layup mapping.",
    explicit_value_list=(
        enum_types_pb2.RosetteSelectionMethod.MINIMUM_DISTANCE,
        enum_types_pb2.RosetteSelectionMethod.MINIMUM_DISTANCE_SUPERPOSED,
    ),
)


@mark_grpc_properties
@register
class LayupMappingObject(CreatableTreeObject, IdTreeObject):
    """Instantiate a layup mapping object.

    Parameters
    ----------
    name :
        Name of the layup mapping object.
    active :
        Inactive layup mapping objects are not used.
    element_technology :
        Determines the element technology used for the layup mapping.
        Can be either ``"layered_element"`` or ``"reinforcing"``.
        Note that only brick and prism elements support the layered option,
        while reinforcing technology can be combined with all types of
        solid elements.
    shell_element_sets :
        Defines the shell area whose modeling plies are mapped onto the
        solid mesh.
        Used only if ``use_imported_plies`` is False.
    use_imported_plies :
        If True, imported modeling plies are mapped onto the solid mesh.
    select_all_plies :
        Determines whether all plies are selected for mapping, or only
        the ones specified in the ``sequences`` parameter.

        * If ``use_imported_plies`` is False and ``select_all_plies`` is True,
          all plies in the selected shell area are selected.
        * If ``use_imported_plies`` is True and ``select_all_plies`` is True,
          all imported modeling plies are selected.

    sequences :
        Determines which plies are considered for mapping. The intersection
        of shell elements and plies are then mapped onto the solid mesh.

        * If ``use_imported_plies`` is False, modeling groups and modeling
          plies are allowed
        * If ``use_imported_plies`` is True, imported modeling groups and
          imported modeling plies are allowed

    entire_solid_mesh :
        If True, the selected layup is mapped onto the entire solid mesh.
        Otherwise, the ``solid_element_sets`` parameter is used to refine
        the scope.
    solid_element_sets :
        List of solid element sets to which the mapping is applied.
    scale_ply_thicknesses :
        If True, scale the layer thickness to fill the gaps within an element,
        instead of filling them with the void material.
        Used only if ``element_technology`` is ``"layered_element"``.
    void_material :
        Material used to fill the gaps with.
        Only used if ``scale_ply_thicknesses`` is False, and
        the ``element_technology`` is ``"layered_element"``.
    minimum_void_material_thickness :
        Only gaps with a thickness greater than this value are filled with
        the void material.
        Only used if ``scale_ply_thicknesses`` is False, and
        the ``element_technology`` is ``"layered_element"``.
    delete_lost_elements :
        If True, elements without layup and degenerated elements are deleted.
        Otherwise, they are filled with the filler material.
        Used only if ``element_technology`` is ``"layered_element"``.
    filler_material :
        Material used to fill the degenerated elements.
        Only used if ``delete_lost_elements`` is False, and
        the ``element_technology`` is ``"layered_element"``.
    rosettes :
        List of rosettes to set the coordinate system of the filler elements.
        Used only if ``element_technology`` is ``"layered_element"``.
    rosette_selection_method :
        Defines how the coordinate systems are applied for the filler elements.
        Used only if ``element_technology`` is ``"layered_element"``.
    reinforcing_behavior :
        Determines whether the reinforcing elements carry tension and / or
        compression loads.
        Used only if ``element_technology`` is ``"reinforcing"``.
    base_element_material_handling :
        Determines how the base material is handled in the area with reinforcing
        elements. Can be either ``"remove"`` or ``"retain"``.
        Used only if ``element_technology`` is ``"reinforcing"``.
    stress_state :
        Specifies the stress state of the reinforcing elements.
        Used only if ``element_technology`` is ``"reinforcing"``.
    base_material :
        Define the initial material of the solid elements which will be reinforced
        by the selected plies.
        Used only if ``element_technology`` is ``"reinforcing"``.
    base_element_rosettes :
        List of rosettes to set the coordinate system of the base elements.
        This is important if the base material has an orthotropic characteristic.
        Used only if ``element_technology`` is ``"reinforcing"``.
    base_element_rosette_selection_method :
        Defines how the coordinate systems are applied for the base elements.
        Used only if ``element_technology`` is ``"reinforcing"``.

    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "layup_mapping_objects"
    _OBJECT_INFO_TYPE = layup_mapping_object_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = layup_mapping_object_pb2.CreateRequest
    _SUPPORTED_SINCE = "25.1"

    def __init__(
        self,
        *,
        name: str = "LayupMappingObject",
        active: bool = True,
        element_technology: ElementTechnology = ElementTechnology.LAYERED_ELEMENT,
        shell_element_sets: Sequence[ElementSet] = (),
        use_imported_plies: bool = False,
        select_all_plies: bool = True,
        sequences: Sequence[
            ModelingGroup | ModelingPly | ImportedModelingGroup | ImportedModelingPly
        ] = (),
        entire_solid_mesh: bool = True,
        solid_element_sets: Sequence[SolidElementSet] = (),
        scale_ply_thicknesses: bool = True,
        void_material: Material | None = None,
        minimum_void_material_thickness: float = 1e-6,
        delete_lost_elements: bool = True,
        filler_material: Material | None = None,
        rosettes: Sequence[Rosette] = (),
        rosette_selection_method: LayupMappingRosetteSelectionMethod = LayupMappingRosetteSelectionMethod.MINIMUM_DISTANCE,  # type: ignore # noqa: E501
        reinforcing_behavior: ReinforcingBehavior = ReinforcingBehavior.TENSION_AND_COMPRESSION,
        base_element_material_handling: BaseElementMaterialHandling = BaseElementMaterialHandling.REMOVE,  # noqa: E501
        stress_state: StressStateType = StressStateType.PLANE_STRESS_STATE_WITH_TRANSVERSE_SHEAR_AND_BENDING_STIFFNESS,  # noqa: E501
        base_material: Material | None = None,
        base_element_rosettes: Sequence[Rosette] = (),
        base_element_rosette_selection_method: LayupMappingRosetteSelectionMethod = LayupMappingRosetteSelectionMethod.MINIMUM_DISTANCE,  # type: ignore # noqa: E501
    ):
        super().__init__(name=name)
        self.active = active
        self.element_technology = element_technology
        self.shell_element_sets = shell_element_sets
        self.use_imported_plies = use_imported_plies
        self.select_all_plies = select_all_plies
        self.sequences = sequences
        self.entire_solid_mesh = entire_solid_mesh
        self.solid_element_sets = solid_element_sets
        self.scale_ply_thicknesses = scale_ply_thicknesses
        self.void_material = void_material
        self.minimum_void_material_thickness = minimum_void_material_thickness
        self.delete_lost_elements = delete_lost_elements
        self.filler_material = filler_material
        self.rosettes = rosettes
        self.rosette_selection_method = rosette_selection_method
        self.reinforcing_behavior = reinforcing_behavior
        self.base_element_material_handling = base_element_material_handling
        self.stress_state = stress_state
        self.base_material = base_material
        self.base_element_rosettes = base_element_rosettes
        self.base_element_rosette_selection_method = base_element_rosette_selection_method

    def _create_stub(self) -> layup_mapping_object_pb2_grpc.ObjectServiceStub:
        return layup_mapping_object_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)

    active: ReadWriteProperty[bool, bool] = grpc_data_property("properties.active")
    element_technology = grpc_data_property(
        "properties.element_technology",
        from_protobuf=element_technology_from_pb,
        to_protobuf=element_technology_to_pb,
    )
    shell_element_sets = define_linked_object_list(
        "properties.shell_element_sets", object_class=ElementSet
    )
    use_imported_plies: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.use_imported_plies"
    )
    select_all_plies: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.select_all_plies"
    )
    sequences = define_polymorphic_linked_object_list(
        "properties.sequences",
        allowed_types=(ModelingGroup, ModelingPly, ImportedModelingGroup, ImportedModelingPly),
    )
    entire_solid_mesh: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.entire_solid_mesh"
    )
    solid_element_sets = define_linked_object_list(
        "properties.solid_element_sets", object_class=SolidElementSet
    )
    scale_ply_thicknesses: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.scale_ply_thicknesses"
    )
    void_material = grpc_link_property("properties.void_material", allowed_types=(Material,))
    minimum_void_material_thickness: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.minimum_void_material_thickness"
    )
    delete_lost_elements: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.delete_lost_elements"
    )
    filler_material = grpc_link_property("properties.filler_material", allowed_types=(Material,))
    rosettes = define_linked_object_list("properties.rosettes", object_class=Rosette)
    rosette_selection_method = grpc_data_property(
        "properties.rosette_selection_method",
        from_protobuf=layup_mapping_rosette_selection_method_from_pb,
        to_protobuf=layup_mapping_rosette_selection_method_to_pb,
    )
    reinforcing_behavior = grpc_data_property(
        "properties.reinforcing_behavior",
        from_protobuf=reinforcing_behavior_from_pb,
        to_protobuf=reinforcing_behavior_to_pb,
    )
    base_element_material_handling = grpc_data_property(
        "properties.base_element_material_handling",
        from_protobuf=base_element_material_handling_from_pb,
        to_protobuf=base_element_material_handling_to_pb,
    )
    stress_state = grpc_data_property(
        "properties.stress_state",
        from_protobuf=stress_state_type_from_pb,
        to_protobuf=stress_state_type_to_pb,
    )
    base_material = grpc_link_property("properties.base_material", allowed_types=(Material,))
    base_element_rosettes = define_linked_object_list(
        "properties.base_element_rosettes", object_class=Rosette
    )
    base_element_rosette_selection_method = grpc_data_property(
        "properties.base_element_rosette_selection_method",
        from_protobuf=layup_mapping_rosette_selection_method_from_pb,
        to_protobuf=layup_mapping_rosette_selection_method_to_pb,
    )

    analysis_plies = get_read_only_collection_property(
        AnalysisPly, analysis_ply_pb2_grpc.ObjectServiceStub, supported_since="25.2"
    )
    imported_analysis_plies = get_read_only_collection_property(
        ImportedAnalysisPly,
        imported_analysis_ply_pb2_grpc.ObjectServiceStub,
        supported_since="25.2",
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Helpers for synchronizing object properties with the backend via gRPC.

Defines helpers which can be used to define properties which are
automatically synchronized with the backend via gRPC.
"""
from __future__ import annotations

from collections.abc import Callable
from functools import reduce
import sys
from typing import TYPE_CHECKING, Any, TypeVar

from google.protobuf.message import Message

from ansys.api.acp.v0.base_pb2 import ResourcePath

from ..._utils.property_protocols import ReadOnlyProperty, ReadWriteProperty
from .polymorphic_from_pb import CreatableFromResourcePath, tree_object_from_resource_path
from .protocols import Editable, GrpcObjectBase, ObjectInfo, Readable
from .supported_since import supported_since as supported_since_decorator

# Note: The typing of the protobuf objects is fairly loose, maybe it could
# be improved. The main challenge is that we do not encode the structure of
# messages in the type system, and they can contain different fundamental
# or protobuf types.
_PROTOBUF_T = Any
_GET_T = TypeVar("_GET_T")
_SET_T = TypeVar("_SET_T")

_TO_PROTOBUF_T = Callable[[_SET_T], _PROTOBUF_T]
_FROM_PROTOBUF_T = Callable[[_PROTOBUF_T], _GET_T]


if TYPE_CHECKING:  # pragma: no cover
    # This is needed because mypy does not understand custom property
    # subclasses.
    # See https://github.com/python/mypy/issues/6158
    _exposed_grpc_property = property
    _exposed_grpc_mapping_property = property
else:

    class _exposed_grpc_property(property):
        """Mark a property as exposed via gRPC.

        Wrapper around 'property', used to signal that the object should
        be collected into the '_GRPC_PROPERTIES' class attribute.
        """

        pass

    class _exposed_grpc_mapping_property(_exposed_grpc_property):
        """Mark a property as a mapping.

        Wrapper around 'property', used to signal that the object should
        be collected into the '_GRPC_PROPERTIES' class attribute.
        """

        def __init__(self, *args, value_type, read_only: bool, **kwargs):
            super().__init__(*args, **kwargs)
            self._value_type = value_type
            self._read_only = read_only


T = TypeVar("T", bound=type[GrpcObjectBase])


def mark_grpc_properties(cls: T) -> T:
    """Class decorator to collect properties marked as exposed via gRPC.

    This decorator should be applied to all classes which define gRPC
    properties.
    """
    props: list[str] = []
    # Loop is needed because we otherwise get only the _GRPC_PROPERTIES of one of the base classes.
    for base_cls in reversed(cls.__bases__):
        if hasattr(base_cls, "_GRPC_PROPERTIES"):
            props.extend(base_cls._GRPC_PROPERTIES)
    for key, value in vars(cls).items():
        if isinstance(value, _exposed_grpc_property):
            props.append(key)
    props_unique = []
    for name in props:
        if name not in props_unique:
            props_unique.append(name)
    cls._GRPC_PROPERTIES = tuple(props_unique)

    # The 'mark_grpc_properties' decorator is also used on intermediate base
    # classes which do not have the '_SUPPORTED_SINCE' attribute. We only want
    # to add the version information to the final class.
    if hasattr(cls, "_SUPPORTED_SINCE"):
        if isinstance(cls.__doc__, str):
            # When adding to the docstring, we need to match the existing
            # indentation of the docstring (except the first line).
            # See PEP 257 'Handling Docstring Indentation'.
            # Alternatively, we could strip the common indentation from the
            # docstring.
            indent = sys.maxsize
            for line in cls.__doc__.splitlines()[1:]:
                stripped = line.lstrip()
                if stripped:  # ignore empty lines
                    indent = min(indent, len(line) - len(stripped))
            if indent == sys.maxsize:
                indent = 0
            cls.__doc__ += (
                f"\n\n{indent * ' '}*Added in ACP server version {cls._SUPPORTED_SINCE}.*\n"
            )
    return cls


def grpc_linked_object_getter(
    name: str, readable_since: str | None = None
) -> Callable[[Readable], Any]:
    """Create a getter method which obtains the linked server object."""

    @supported_since_decorator(
        readable_since,
        # The default error message uses 'inner' as the method name, which is confusing
        err_msg_tpl=(
            f"The property '{name.split('.')[-1]}' is only readable since version {{required_version}} "
            f"of the ACP gRPC server. The current server version is {{server_version}}."
        ),
    )
    def inner(self: Readable) -> CreatableFromResourcePath | None:
        if not self._is_stored:
            raise RuntimeError(f"Cannot get linked object '{name}' from unstored object")
        self._get()
        object_resource_path = _get_data_attribute(self._pb_object, name)

        return tree_object_from_resource_path(
            object_resource_path, server_wrapper=self._server_wrapper
        )

    return inner


def _get_data_attribute(pb_obj: Message, name: str, check_optional: bool = False) -> _PROTOBUF_T:
    name_parts = name.split(".")
    if check_optional:
        parent_obj = reduce(getattr, name_parts[:-1], pb_obj)
        try:
            if hasattr(parent_obj, "HasField") and not parent_obj.HasField(name_parts[-1]):
                return None
        except ValueError:
            # The field is not optional, so HasField raises a ValueError.
            pass
    return reduce(getattr, name_parts, pb_obj)


def grpc_data_getter(
    name: str,
    from_protobuf: _FROM_PROTOBUF_T[_GET_T],
    check_optional: bool = False,
    getter_func: Callable[[Message, str, bool], _PROTOBUF_T] = _get_data_attribute,
    supported_since: str | None = None,
) -> Callable[[Readable], _GET_T]:
    """Create a getter method which obtains the server object via the gRPC Get endpoint.

    Parameters
    ----------
    from_protobuf :
        Function to convert the protobuf object to the type exposed by the
        property.
    check_optional :
        If ``True``, the getter will return ``None`` if the property is not
        set on the protobuf object. Otherwise, the default protobuf value
        will be used.
    """

    @supported_since_decorator(
        supported_since,
        # The default error message uses 'inner' as the method name, which is confusing
        err_msg_tpl=(
            f"The property '{name.split('.')[-1]}' is only readable since version {{required_version}} "
            f"of the ACP gRPC server. The current server version is {{server_version}}."
        ),
    )
    def inner(self: Readable) -> Any:
        self._get_if_stored()
        pb_attribute = getter_func(self._pb_object, name, check_optional)
        if check_optional and pb_attribute is None:
            return None
        return from_protobuf(pb_attribute)

    return inner


def grpc_linked_object_setter(
    name: str, to_protobuf: _TO_PROTOBUF_T[Readable | None], writable_since: str | None = None
) -> Callable[[Editable, Readable | None], None]:
    """Create a setter method which updates the linked object via the gRPC Put endpoint."""
    func = grpc_data_setter(name=name, to_protobuf=to_protobuf, supported_since=writable_since)

    def inner(self: Editable, value: Readable | None) -> None:
        if value is not None and not value._is_stored:
            raise Exception("Cannot link to an unstored object.")
        func(self, value)

    return inner


def _set_data_attribute(pb_obj: ObjectInfo, name: str, value: _PROTOBUF_T) -> None:
    name_parts = name.split(".")

    try:
        parent = reduce(getattr, name_parts[:-1], pb_obj)
        setattr(parent, name_parts[-1], value)
    except AttributeError:
        target_object: Any = reduce(getattr, name_parts, pb_obj)
        if hasattr(target_object, "CopyFrom"):
            target_object.CopyFrom(value)
        else:
            try:
                target_object[:] = value
            except TypeError:
                del target_object[:]
                for item in value:
                    target_object.add().CopyFrom(item)


def grpc_data_setter(
    name: str,
    to_protobuf: _TO_PROTOBUF_T[_SET_T],
    setter_func: Callable[[ObjectInfo, str, _PROTOBUF_T], None] = _set_data_attribute,
    supported_since: str | None = None,
) -> Callable[[Editable, _SET_T], None]:
    """Create a setter method which updates the server object via the gRPC Put endpoint."""

    @supported_since_decorator(
        supported_since,
        # The default error message uses 'inner' as the method name, which is confusing
        err_msg_tpl=(
            f"The property '{name.split('.')[-1]}' is only editable since version {{required_version}} "
            f"of the ACP gRPC server. The current server version is {{server_version}}."
        ),
    )
    def inner(self: Editable, value: _SET_T) -> None:
        self._get_if_stored()
        current_value = _get_data_attribute(self._pb_object, name, check_optional=True)
        value_pb = to_protobuf(value)
        try:
            needs_updating = current_value != value_pb
        except TypeError:
            needs_updating = True
        if needs_updating:
            setter_func(self._pb_object, name, value_pb)
            self._put_if_stored()

    return inner


AnyT = TypeVar("AnyT")


def _wrap_doc(obj: AnyT, doc: str | None) -> AnyT:
    if doc is not None:
        obj.__doc__ = doc
    return obj


def grpc_data_property(
    name: str,
    to_protobuf: _TO_PROTOBUF_T[_SET_T] = lambda x: x,
    from_protobuf: _FROM_PROTOBUF_T[_GET_T] = lambda x: x,
    check_optional: bool = False,
    doc: str | None = None,
    setter_func: Callable[[ObjectInfo, str, _PROTOBUF_T], None] = _set_data_attribute,
    getter_func: Callable[[Message, str, bool], _PROTOBUF_T] = _get_data_attribute,
    readable_since: str | None = None,
    writable_since: str | None = None,
) -> ReadWriteProperty[_GET_T, _SET_T]:
    """Define a property which is synchronized with the backend via gRPC.

    Helper for defining properties accessed via gRPC. The property getter
    and setter make calls to the gRPC Get and Put endpoints to synchronize
    the local object with the remote backend.

    Parameters
    ----------
    name :
        Name of the property.
    to_protobuf :
        Function to convert the property value to the protobuf type.
    from_protobuf :
        Function to convert the protobuf object to the type exposed by the
        property.
    check_optional :
        If ``True``, the getter will return ``None`` if the property is not
        set on the protobuf object. Otherwise, the default protobuf value
        will be used.
    doc :
        Docstring for the property.
    setter_func :
        Function to set the property value. Can be customized to
        implement additional checks or in case properties depend
        on each other.
    getter_func :
        Function to get the property value. Can be customized to
        implement additional checks or in case properties depend
        on each other
    readable_since :
        Version since which the property is supported for reading.
    writable_since :
        Version since which the property is supported for setting.
    """
    # Note jvonrick August 2023: We don't ensure with typechecks that the property returned here is
    # compatible with the class on which this property is created. For example:
    # grpc_data_setter returns a callable that expects an editable object as the first argument.
    # But this property can also be added to a class that does not satisfy the Editable
    # Protocol
    # See the discussion here on why it is hard to have typed properties:
    # https://github.com/python/typing/issues/985
    return _wrap_doc(
        _exposed_grpc_property(
            grpc_data_getter(
                name,
                from_protobuf=from_protobuf,
                check_optional=check_optional,
                getter_func=getter_func,
                supported_since=readable_since,
            )
        ).setter(
            grpc_data_setter(
                name,
                to_protobuf=to_protobuf,
                setter_func=setter_func,
                supported_since=writable_since,
            )
        ),
        doc=doc,
    )


def grpc_data_property_read_only(
    name: str,
    from_protobuf: _FROM_PROTOBUF_T[_GET_T] = lambda x: x,
    check_optional: bool = False,
    doc: str | None = None,
    supported_since: str | None = None,
) -> ReadOnlyProperty[_GET_T]:
    """Define a read-only property which is synchronized with the backend via gRPC.

    Helper for defining properties accessed via gRPC. The property getter
    makes call to the gRPC Get endpoints to synchronize
    the local object with the remote backend.

    Parameters
    ----------
    name :
        Name of the property.
    from_protobuf :
        Function to convert the protobuf object to the type exposed by the
        property.
    check_optional :
        If ``True``, the getter will return ``None`` if the property is not
        set on the protobuf object. Otherwise, the default protobuf value
        will be used.
    doc :
        Docstring for the property.
    supported_since :
        Version since which the property is supported.
    """
    return _wrap_doc(
        _exposed_grpc_property(
            grpc_data_getter(
                name,
                from_protobuf=from_protobuf,
                check_optional=check_optional,
                supported_since=supported_since,
            )
        ),
        doc=doc,
    )


def grpc_link_property(
    name: str,
    *,
    doc: str | None = None,
    allowed_types: type[GrpcObjectBase] | tuple[type[GrpcObjectBase], ...],
    readable_since: str | None = None,
    writable_since: str | None = None,
) -> Any:
    """Define a gRPC-backed property linking to another object.

    Helper for defining linked properties accessed via gRPC. The property getter
    makes call to the gRPC Get endpoints to get the linked object

    Parameters
    ----------
    name :
        Name of the property.
    doc :
        Docstring for the property.
    allowed_types :
        Types which are allowed to be set on the property. An
        error will be raised if an object of a different type is set.
    readable_since :
        Version since which the property is supported for reading.
    writable_since :
        Version since which the property is supported for setting.
    """

    def to_protobuf(obj: Readable | None) -> ResourcePath:
        if obj is None:
            return ResourcePath(value="")
        if not isinstance(obj, allowed_types):
            name_part = name.split(".")[-1]
            raise TypeError(
                f"Cannot set '{name_part}': Expected object of type {allowed_types}, "
                f"got type {type(obj)} instead. Given object: {obj}"
            )
        return obj._resource_path

    return _wrap_doc(
        _exposed_grpc_property(
            grpc_linked_object_getter(name=name, readable_since=readable_since)
        ).setter(
            # Resource path represents an object that is not set as an empty string
            grpc_linked_object_setter(
                name=name, to_protobuf=to_protobuf, writable_since=writable_since
            )
        ),
        doc=doc,
    )


def grpc_link_property_read_only(name: str, doc: str | None = None) -> Any:
    """Define a read-only gRPC-backed property linking to another object.

    Helper for defining linked properties accessed via gRPC. The property getter
    makes call to the gRPC Get endpoints to get the linked object

    Parameters
    ----------
    name :
        Name of the property.
    doc :
        Docstring for the property.
    """
    return _wrap_doc(_exposed_grpc_property(grpc_linked_object_getter(name)), doc=doc)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections.abc import Iterable, Iterator
from dataclasses import dataclass

import networkx as nx

from ._tree_objects._grpc_helpers.linked_object_helpers import get_linked_paths
from ._tree_objects._grpc_helpers.mapping import Mapping
from ._tree_objects._grpc_helpers.polymorphic_from_pb import tree_object_from_resource_path
from ._tree_objects.base import CreatableTreeObject, TreeObject


@dataclass
class _WalkTreeOptions:
    include_children: bool
    include_linked_objects: bool


def _build_dependency_graph(
    *, source_objects: Iterable[CreatableTreeObject], options: _WalkTreeOptions
) -> nx.DiGraph:
    """Build a dependency graph of the given objects."""
    graph = nx.DiGraph()

    # We need to manually keep track of which objects have been visited,
    # since the node may also be created when being linked to.
    visited_objects: set[CreatableTreeObject] = set()
    for tree_object in source_objects:
        _build_dependency_graph_impl(
            tree_object=tree_object, graph=graph, visited_objects=visited_objects, options=options
        )
    return graph


def _build_dependency_graph_impl(
    *,
    tree_object: CreatableTreeObject,
    graph: nx.DiGraph,
    visited_objects: set[CreatableTreeObject],
    options: _WalkTreeOptions,
) -> None:

    if tree_object in visited_objects:
        return

    visited_objects.add(tree_object)
    graph.add_node(tree_object)

    if options.include_children:
        for child_object in _yield_child_objects(tree_object):
            if not isinstance(child_object, CreatableTreeObject):
                continue
            graph.add_edge(child_object, tree_object)
            _build_dependency_graph_impl(
                tree_object=child_object,
                graph=graph,
                visited_objects=visited_objects,
                options=options,
            )
    if options.include_linked_objects:
        for linked_object in _yield_linked_objects(tree_object):
            graph.add_edge(tree_object, linked_object)
            _build_dependency_graph_impl(
                tree_object=linked_object,
                graph=graph,
                visited_objects=visited_objects,
                options=options,
            )


def _yield_child_objects(tree_object: TreeObject) -> Iterator[TreeObject]:
    for attr_name in tree_object._GRPC_PROPERTIES:
        try:
            attr = getattr(tree_object, attr_name)
        except (AttributeError, RuntimeError):
            continue
        if isinstance(attr, Mapping):
            yield from attr.values()


def _yield_linked_objects(tree_object: TreeObject) -> Iterator[CreatableTreeObject]:
    for linked_path in get_linked_paths(tree_object._pb_object.properties):
        linked_object = tree_object_from_resource_path(
            linked_path, server_wrapper=tree_object._server_wrapper
        )
        assert isinstance(linked_object, CreatableTreeObject)
        yield linked_object



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Protocol

import grpc

from .._utils.typing_helper import StrEnum

__all__ = ["LaunchMode"]


class ServerKey(StrEnum):
    MAIN = "main"
    FILE_TRANSFER = "file_transfer"


class LaunchMode(StrEnum):
    """Available launch modes for ACP."""

    DIRECT = "direct"
    DOCKER_COMPOSE = "docker_compose"
    CONNECT = "connect"


class ServerProtocol(Protocol):
    """Interface definition for ACP gRPC servers."""

    @property
    def channels(self) -> dict[str, grpc.Channel]: ...

    def check(self, timeout: float | None = None) -> bool: ...

    def wait(self, timeout: float) -> None: ...


class ControllableServerProtocol(ServerProtocol, Protocol):
    """Interface definition for ACP servers which can be remotely started / stopped."""

    def start(self) -> None: ...

    def stop(self, *, timeout: float | None = None) -> None: ...

    def restart(self, *, stop_timeout: float | None = None) -> None: ...



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable
import textwrap
import typing
from typing import Any, Protocol

from google.protobuf.message import Message
import grpc
from packaging.version import Version

from ansys.api.acp.v0.base_pb2 import (
    BasicInfo,
    CollectionPath,
    DeleteRequest,
    Empty,
    GetRequest,
    ListRequest,
    ResourcePath,
)

if typing.TYPE_CHECKING:  # pragma: no cover
    from ..base import ServerWrapper


class CreateRequest(Protocol):
    """Interface definition for CreateRequest messages.

    The CreateRequest message is used to create a new object in a collection.
    """

    def __init__(self, collection_path: CollectionPath, name: str, properties: Message): ...


class ObjectInfo(Protocol):
    """Interface definition for ObjectInfo messages.

    The ObjectInfo message contains the full information about an object.
    """

    @property
    def info(self) -> BasicInfo:
        """Provide basic information about the object."""
        ...

    @property
    def properties(self) -> Message:
        """Properties of the object."""
        ...


class ListReply(Protocol):
    """Interface definition for ListReply messages.

    The ListReply message contains a list of objects in a collection.
    """

    @property
    def objects(self) -> list[ObjectInfo]:
        """List of objects in the collection."""
        ...


class EditableResourceStub(Protocol):
    """Interface definition for ACP Resource service stubs.

    This interface defines the edit methods for ACP Resource service stubs.
    """

    def Put(self, request: ObjectInfo) -> ObjectInfo:
        """RPC method for updating an object."""
        ...

    def Delete(self, request: DeleteRequest) -> Empty:
        """RPC method for deleting an object."""
        ...


class ReadableResourceStub(Protocol):
    """Interface definition for ACP Resource service stubs.

    This interface defines the read methods for ACP Resource service stubs.
    """

    def __init__(self, channel: grpc.Channel): ...

    def Get(self, request: GetRequest) -> ObjectInfo:
        """RPC method for getting an object's information."""
        ...

    def List(self, request: ListRequest) -> ListReply:
        """RPC method for listing objects in a collection."""
        ...


class EditableAndReadableResourceStub(EditableResourceStub, ReadableResourceStub, Protocol):
    """Interface definition for ACP Resource service stubs.

    This interface defines the edit and read methods for ACP Resource service stubs.
    """

    ...


class CreatableResourceStub(Protocol):
    """Interface definition for ACP Resource service stubs.

    This interface defines the create methods for ACP Resource service stubs.
    """

    def Create(self, request: CreateRequest) -> ObjectInfo:
        """RPC method for creating an object."""
        ...


class CreatableEditableAndReadableResourceStub(
    CreatableResourceStub, EditableResourceStub, ReadableResourceStub, Protocol
):
    """Interface definition for ACP Resource service stubs.

    This interface defines the create, edit, and read methods for ACP Resource service stubs.
    """

    ...


class GrpcObjectBase(Protocol):
    """Interface definition for objects which are backed by a gRPC API."""

    __slots__: Iterable[str] = tuple()
    _GRPC_PROPERTIES: tuple[str, ...] = tuple()
    _SUPPORTED_SINCE: str

    def __str__(self) -> str:
        string_items = []
        for attr_name in self._GRPC_PROPERTIES:
            try:
                value_repr = repr(getattr(self, attr_name))
            except:
                value_repr = "<unavailable>"
            string_items.append(f"{attr_name}={value_repr}")
        type_name = type(self).__name__
        if not string_items:
            content = ""
        elif len(string_items) == 1:
            content = string_items[0]
        else:
            content = ",\n".join(string_items)
            content = f"\n{textwrap.indent(content, ' ' * 4)}\n"
        return f"{type_name}({content})"


class Readable(Protocol):
    """Interface definition for readable objects."""

    def _get(self) -> None: ...

    def _get_if_stored(self) -> None: ...

    @property
    def _is_stored(self) -> bool: ...

    @property
    def _server_wrapper(self) -> ServerWrapper: ...

    @property
    def _resource_path(self) -> ResourcePath: ...

    _pb_object: Any

    @property
    def _server_version(self) -> Version | None: ...


class Editable(Readable, Protocol):
    """Interface definition for editable objects."""

    def _put(self) -> None: ...

    def _put_if_stored(self) -> None: ...



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Callable, Container, Iterable
import dataclasses
from typing import Any

import numpy as np
from typing_extensions import Self

from ansys.api.acp.v0 import modeling_ply_pb2, modeling_ply_pb2_grpc, production_ply_pb2_grpc

from .._utils.array_conversions import to_1D_double_array, to_tuple_from_1D_array
from .._utils.property_protocols import ReadWriteProperty
from ._elemental_or_nodal_data import (
    ElementalData,
    NodalData,
    ScalarData,
    VectorData,
    elemental_data_property,
    nodal_data_property,
)
from ._grpc_helpers.edge_property_list import (
    GenericEdgePropertyType,
    define_add_method,
    define_edge_property_list,
)
from ._grpc_helpers.linked_object_list import define_linked_object_list
from ._grpc_helpers.mapping import get_read_only_collection_property
from ._grpc_helpers.property_helper import (
    _exposed_grpc_property,
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from ._mesh_data import full_mesh_property, shell_mesh_property
from .base import CreatableTreeObject, IdTreeObject
from .edge_set import EdgeSet
from .enums import (
    DrapingType,
    ThicknessFieldType,
    ThicknessType,
    draping_type_from_pb,
    draping_type_to_pb,
    status_type_from_pb,
    thickness_field_type_from_pb,
    thickness_field_type_to_pb,
    thickness_type_from_pb,
    thickness_type_to_pb,
)
from .fabric import Fabric
from .linked_selection_rule import LinkedSelectionRule
from .lookup_table_1d_column import LookUpTable1DColumn
from .lookup_table_3d_column import LookUpTable3DColumn
from .object_registry import register
from .oriented_selection_set import OrientedSelectionSet
from .production_ply import ProductionPly
from .stackup import Stackup
from .sublaminate import SubLaminate
from .virtual_geometry import VirtualGeometry

__all__ = ["ModelingPly", "ModelingPlyElementalData", "ModelingPlyNodalData", "TaperEdge"]


@dataclasses.dataclass
class ModelingPlyElementalData(ElementalData):
    """Represents elemental data for a Modeling Ply."""

    normal: VectorData | None = None
    orientation: VectorData | None = None
    reference_direction: VectorData | None = None
    fiber_direction: VectorData | None = None
    draped_fiber_direction: VectorData | None = None
    transverse_direction: VectorData | None = None
    draped_transverse_direction: VectorData | None = None
    thickness: ScalarData[np.float64] | None = None
    relative_thickness_correction: ScalarData[np.float64] | None = None
    design_angle: ScalarData[np.float64] | None = None
    shear_angle: ScalarData[np.float64] | None = None
    draped_fiber_angle: ScalarData[np.float64] | None = None
    draped_transverse_angle: ScalarData[np.float64] | None = None
    area: ScalarData[np.float64] | None = None
    price: ScalarData[np.float64] | None = None
    volume: ScalarData[np.float64] | None = None
    mass: ScalarData[np.float64] | None = None
    offset: ScalarData[np.float64] | None = None
    cog: VectorData | None = None


@dataclasses.dataclass
class ModelingPlyNodalData(NodalData):
    """Represents nodal data for a Modeling Ply."""

    ply_offset: VectorData | None = None


@mark_grpc_properties
class TaperEdge(GenericEdgePropertyType):
    """Defines a taper edge.

    Parameters
    ----------
    edge_set :
        Defines the edge along which the ply tapering is applied.
    angle :
        Defines the angle between the cutting plane and  the reference surface.
    offset :
        Moves the cutting plane along the out-of-plane direction.
        A negative value cuts the elements at the edge where the in-plane
        offset is ``-offset/tan(angle)``.
    """

    _SUPPORTED_SINCE = "24.2"

    def __init__(self, edge_set: EdgeSet, *, angle: float, offset: float):
        self._callback_apply_changes: Callable[[], None] | None = None
        self.edge_set = edge_set
        self.angle = angle
        self.offset = offset

    @_exposed_grpc_property
    def edge_set(self) -> EdgeSet:
        """Edge along which the ply tapering is applied."""
        return self._edge_set

    @edge_set.setter
    def edge_set(self, edge_set: EdgeSet) -> None:
        if not isinstance(edge_set, EdgeSet):
            raise TypeError(f"Expected an EdgeSet, got {type(edge_set)}.")
        self._edge_set = edge_set
        if self._callback_apply_changes is not None:
            self._callback_apply_changes()

    @_exposed_grpc_property
    def angle(self) -> float:
        """Angle between the cutting plane and  the reference surface."""
        return self._angle

    @angle.setter
    def angle(self, angle: float) -> None:
        self._angle = angle
        if self._callback_apply_changes is not None:
            self._callback_apply_changes()

    @_exposed_grpc_property
    def offset(self) -> float:
        """Move the cutting plane along the out-of-plane direction.

        A negative value cuts the elements at the edge where the in-plane
        offset is ``-offset/tan(angle)``.
        """
        return self._offset

    @offset.setter
    def offset(self, offset: float) -> None:
        self._offset = offset
        if self._callback_apply_changes is not None:
            self._callback_apply_changes()

    def _set_callback_apply_changes(self, callback_apply_changes: Callable[[], None]) -> None:
        self._callback_apply_changes = callback_apply_changes

    @classmethod
    def _from_pb_object(
        cls,
        parent_object: CreatableTreeObject,
        message: modeling_ply_pb2.TaperEdge,
        apply_changes: Callable[[], None],
    ) -> Self:
        edge_set = EdgeSet._from_resource_path(
            resource_path=message.edge_set, server_wrapper=parent_object._server_wrapper
        )

        new_obj = cls(
            edge_set=edge_set,
            angle=message.angle,
            offset=message.offset,
        )
        new_obj._set_callback_apply_changes(apply_changes)
        return new_obj

    def _to_pb_object(self) -> modeling_ply_pb2.TaperEdge:
        return modeling_ply_pb2.TaperEdge(
            edge_set=self.edge_set._resource_path,
            angle=self.angle,
            offset=self.offset,
        )

    def _check(self) -> bool:
        return bool(self.edge_set._resource_path.value)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, TaperEdge):
            return (
                self.edge_set == other.edge_set
                and self.angle == other.angle
                and self.offset == other.offset
            )
        return False

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(edge_set={self.edge_set!r}, "
            f"angle={self.angle!r}, offset={self.offset!r})"
        )

    def clone(self) -> Self:
        """Create a new unstored TaperEdge with the same properties."""
        return type(self)(
            edge_set=self.edge_set,
            angle=self.angle,
            offset=self.offset,
        )


@mark_grpc_properties
@register
class ModelingPly(CreatableTreeObject, IdTreeObject):
    """Instantiate a Modeling Ply.

    Parameters
    ----------
    name :
        The name of the Modeling Ply
    ply_material :
        The material (fabric, stackup or sub-laminate) of the ply.
    ply_angle :
        Design angle between the reference direction and the ply fiber direction.
    number_of_layers :
        Number of times the plies are generated.
    active :
        Inactive plies are ignored in ACP and the downstream analysis.
    global_ply_nr :
        Defines the global ply order.
    selection_rules :
        Selection Rules which may limit the extent of the ply.
    draping_type :
        Chooses between different draping formulations.
    draping_seed_point :
        Starting point of the draping algorithm.
    auto_draping_direction :
        If ``True``, the fiber direction of the production ply at the draping
         seed point is used as draping direction.
    draping_direction :
        Set the primary draping direction for the draping algorithm. Only used if
        ``auto_draping_direction`` is ``False``.
    use_default_draping_mesh_size :
        Whether to use the average element size of the shell mesh for the draping.
    draping_mesh_size :
        Defines the mesh size for the draping algorithm.  If set to ``-1.``, the
        mesh size is automatically determined based on the average element size.
    draping_thickness_correction :
        Enables the thickness correction of draped plies based on the draping
        shear angle.
    draping_angle_1_field :
        Correction angle between the fiber and draped fiber directions, in degree.
    draping_angle_2_field :
        Correction angle between the transverse and draped transverse directions,
        in degree. Optional, uses the same values as ``draping_angle_1_field``
        (no shear) by default.
    thickness_type :
        Choose between :attr:`ThicknessType.FROM_GEOMETRY` or
        :attr:`ThicknessType.FROM_TABLE` to define a ply with variable thickness.
        The default value is :attr:`ThicknessType.NOMINAL`, which means the ply
        thickness is constant and determined by the thickness of the ply material.
    thickness_geometry :
        Defines the geometry used to determine the ply thickness. Only applies if
        ``thickness_type`` is :attr:`ThicknessType.FROM_GEOMETRY`.
    thickness_field :
        Defines the look-up table column used to determine the ply thickness.
        Only applies if ``thickness_type`` is :attr:`ThicknessType.FROM_TABLE`.
    thickness_field_type :
        If ``thickness_type`` is :attr:`ThicknessType.FROM_TABLE`, this parameter
        determines how the thickness values are interpreted. They can be either
        absolute values (:attr:`ThicknessFieldType.ABSOLUTE_VALUES`) or relative
        values (:attr:`ThicknessFieldType.RELATIVE_SCALING_FACTOR`).
    taper_edges :
        Defines the taper edges of the ply.
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "modeling_plies"
    _OBJECT_INFO_TYPE = modeling_ply_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = modeling_ply_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "ModelingPly",
        ply_material: Fabric | Stackup | SubLaminate | None = None,
        oriented_selection_sets: Container[OrientedSelectionSet] = (),
        ply_angle: float = 0.0,
        number_of_layers: int = 1,
        active: bool = True,
        # Backend will automatically assign a consistent global_ply_nr
        # if global_ply_nr == 0
        global_ply_nr: int = 0,
        selection_rules: Iterable[LinkedSelectionRule] = (),
        draping_type: DrapingType = DrapingType.NO_DRAPING,
        draping_seed_point: tuple[float, float, float] = (0.0, 0.0, 0.0),
        auto_draping_direction: bool = True,
        draping_direction: tuple[float, float, float] = (1.0, 0.0, 0.0),
        use_default_draping_mesh_size: bool = True,
        draping_mesh_size: float = 0.0,
        draping_thickness_correction: bool = True,
        draping_angle_1_field: LookUpTable1DColumn | LookUpTable3DColumn | None = None,
        draping_angle_2_field: LookUpTable1DColumn | LookUpTable3DColumn | None = None,
        thickness_type: ThicknessType = ThicknessType.NOMINAL,
        thickness_geometry: VirtualGeometry | None = None,
        thickness_field: LookUpTable1DColumn | LookUpTable3DColumn | None = None,
        thickness_field_type: ThicknessFieldType = ThicknessFieldType.ABSOLUTE_VALUES,
        taper_edges: Iterable[TaperEdge] = (),
    ):
        super().__init__(name=name)

        self.oriented_selection_sets = oriented_selection_sets
        self.ply_material = ply_material
        self.ply_angle = ply_angle
        self.number_of_layers = number_of_layers
        self.active = active
        self.global_ply_nr = global_ply_nr
        self.selection_rules = selection_rules
        self.draping_type = draping_type
        self.draping_seed_point = draping_seed_point
        self.auto_draping_direction = auto_draping_direction
        self.draping_direction = draping_direction
        self.use_default_draping_mesh_size = use_default_draping_mesh_size
        self.draping_mesh_size = draping_mesh_size
        self.draping_thickness_correction = draping_thickness_correction
        self.draping_angle_1_field = draping_angle_1_field
        self.draping_angle_2_field = draping_angle_2_field
        self.thickness_type = thickness_type
        self.thickness_geometry = thickness_geometry
        self.thickness_field = thickness_field
        self.thickness_field_type = thickness_field_type
        self.taper_edges = taper_edges

    def _create_stub(self) -> modeling_ply_pb2_grpc.ObjectServiceStub:
        return modeling_ply_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)

    ply_material = grpc_link_property(
        "properties.ply_material", allowed_types=(Fabric, Stackup, SubLaminate)
    )

    oriented_selection_sets = define_linked_object_list(
        "properties.oriented_selection_sets", OrientedSelectionSet
    )

    ply_angle: ReadWriteProperty[float, float] = grpc_data_property("properties.ply_angle")
    number_of_layers: ReadWriteProperty[int, int] = grpc_data_property(
        "properties.number_of_layers"
    )
    active: ReadWriteProperty[bool, bool] = grpc_data_property("properties.active")
    global_ply_nr: ReadWriteProperty[int, int] = grpc_data_property("properties.global_ply_nr")

    draping_type = grpc_data_property(
        "properties.draping", from_protobuf=draping_type_from_pb, to_protobuf=draping_type_to_pb
    )
    draping_seed_point = grpc_data_property(
        "properties.draping_seed_point",
        from_protobuf=to_tuple_from_1D_array,
        to_protobuf=to_1D_double_array,
    )
    auto_draping_direction: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.auto_draping_direction"
    )
    draping_direction = grpc_data_property(
        "properties.draping_direction",
        from_protobuf=to_tuple_from_1D_array,
        to_protobuf=to_1D_double_array,
    )
    use_default_draping_mesh_size: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.use_default_draping_mesh_size"
    )
    draping_mesh_size: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.draping_mesh_size"
    )
    draping_thickness_correction: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.draping_thickness_correction"
    )
    draping_angle_1_field = grpc_link_property(
        "properties.draping_angle_1_field", allowed_types=(LookUpTable1DColumn, LookUpTable3DColumn)
    )
    draping_angle_2_field = grpc_link_property(
        "properties.draping_angle_2_field", allowed_types=(LookUpTable1DColumn, LookUpTable3DColumn)
    )

    selection_rules = define_edge_property_list("properties.selection_rules", LinkedSelectionRule)
    add_selection_rule = define_add_method(
        LinkedSelectionRule,
        attribute_name="selection_rules",
        func_name="add_selection_rule",
        parent_class_name="ModelingPly",
        module_name=__module__,
    )

    thickness_type = grpc_data_property(
        "properties.thickness_type",
        from_protobuf=thickness_type_from_pb,
        to_protobuf=thickness_type_to_pb,
    )
    thickness_geometry = grpc_link_property(
        "properties.thickness_geometry", allowed_types=VirtualGeometry
    )
    thickness_field = grpc_link_property(
        "properties.thickness_field", allowed_types=(LookUpTable1DColumn, LookUpTable3DColumn)
    )
    thickness_field_type = grpc_data_property(
        "properties.thickness_field_type",
        from_protobuf=thickness_field_type_from_pb,
        to_protobuf=thickness_field_type_to_pb,
    )

    taper_edges = define_edge_property_list("properties.taper_edges", TaperEdge)
    add_taper_edge = define_add_method(
        TaperEdge,
        attribute_name="taper_edges",
        func_name="add_taper_edge",
        parent_class_name="ModelingPly",
        module_name=__module__,
    )

    production_plies = get_read_only_collection_property(
        ProductionPly, production_ply_pb2_grpc.ObjectServiceStub
    )

    mesh = full_mesh_property
    shell_mesh = shell_mesh_property

    elemental_data = elemental_data_property(ModelingPlyElementalData)
    nodal_data = nodal_data_property(ModelingPlyNodalData)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import dataclasses
import os
import pathlib
import subprocess  # nosec B404
from typing import TextIO

import grpc

from ansys.tools.local_product_launcher.helpers.grpc import check_grpc_health
from ansys.tools.local_product_launcher.helpers.ports import find_free_ports
from ansys.tools.local_product_launcher.interface import (
    METADATA_KEY_DOC,
    LauncherProtocol,
    ServerType,
)
from ansys.tools.path import get_latest_ansys_installation

from .common import ServerKey

__all__ = ["DirectLaunchConfig"]


def _get_default_binary_path() -> str:
    try:
        _, ans_root = get_latest_ansys_installation()
        binary_path = os.path.join(ans_root, "ACP", "acp_grpcserver")
        if os.name == "nt":
            binary_path += ".exe"
        return binary_path
    except (ValueError, FileNotFoundError):
        return ""


@dataclasses.dataclass
class DirectLaunchConfig:
    """Configuration options for launching ACP as a sub-process."""

    binary_path: str = dataclasses.field(
        default=_get_default_binary_path(),
        metadata={METADATA_KEY_DOC: "Path to the ACP gRPC server executable."},
    )
    stdout_file: str = dataclasses.field(
        default=os.devnull,
        metadata={METADATA_KEY_DOC: "File in which the server stdout is stored."},
    )
    stderr_file: str = dataclasses.field(
        default=os.devnull,
        metadata={METADATA_KEY_DOC: "File in which the server stderr is stored."},
    )


class DirectLauncher(LauncherProtocol[DirectLaunchConfig]):
    CONFIG_MODEL = DirectLaunchConfig
    SERVER_SPEC = {ServerKey.MAIN: ServerType.GRPC}

    def __init__(self, *, config: DirectLaunchConfig):
        self._config = config
        self._url: str
        self._process: subprocess.Popen[str] | None = None
        self._stdout: TextIO
        self._stderr: TextIO

    def start(self) -> None:
        # TODO: implement patterns
        stdout_file = self._config.stdout_file
        stderr_file = self._config.stderr_file

        binary = pathlib.Path(self._config.binary_path)
        if not binary.exists():
            raise FileNotFoundError(f"Binary not found: '{binary}'")

        port = find_free_ports()[0]
        self._url = f"localhost:{port}"
        self._stdout = open(stdout_file, mode="w", encoding="utf-8")
        self._stderr = open(stderr_file, mode="w", encoding="utf-8")
        self._process = subprocess.Popen(  # nosec B603: documented in 'security_considerations.rst'
            [
                self._config.binary_path,
                f"--server-address=0.0.0.0:{port}",
            ],
            stdout=self._stdout,
            stderr=self._stderr,
            text=True,
        )

    def stop(self, *, timeout: float | None = None) -> None:
        if self._process is None:
            # The process has not been started, and therefore doesn't need to be stopped
            return
        self._process.terminate()
        try:
            self._process.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            self._process.kill()
            self._process.wait()
        self._stdout.close()
        self._stderr.close()

    def check(self, timeout: float | None = None) -> bool:
        channel = grpc.insecure_channel(self.urls[ServerKey.MAIN])
        return check_grpc_health(channel=channel, timeout=timeout)

    @property
    def urls(self) -> dict[str, str]:
        return {ServerKey.MAIN: self._url}



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""PyACP: Ansys Composite PrepPost (ACP) module for Python.

PyACP enables you to design and analyze layered composite structures.
"""

import importlib.metadata

from . import (
    dpf_integration_helpers,
    extras,
    material_property_sets,
    mechanical_integration_helpers,
    mesh_data,
)
from ._model_printer import get_model_tree, print_model
from ._plotter import get_directions_plotter
from ._recursive_copy import LinkedObjectHandling, recursive_copy
from ._server import (
    ACPInstance,
    ConnectLaunchConfig,
    DirectLaunchConfig,
    DockerComposeLaunchConfig,
    LaunchMode,
    launch_acp,
)
from ._tree_objects import (
    AnalysisPly,
    ArrowType,
    BaseElementMaterialHandling,
    BooleanOperationType,
    BooleanSelectionRule,
    ButtJointSequence,
    CADComponent,
    CADGeometry,
    CoordinateTransformation,
    CutOffGeometry,
    CutOffGeometryOrientationType,
    CutOffMaterialHandling,
    CutOffRuleType,
    CutOffSelectionRule,
    CylindricalSelectionRule,
    DrapingMaterialModel,
    DrapingType,
    DropOffMaterialHandling,
    DropOffSettings,
    DropOffType,
    EdgeSet,
    EdgeSetType,
    ElementalDataType,
    ElementSet,
    ElementTechnology,
    ExtrusionGuide,
    ExtrusionGuideType,
    ExtrusionMethod,
    ExtrusionType,
    Fabric,
    FabricWithAngle,
    FeFormat,
    FieldDefinition,
    GeometricalRuleType,
    GeometricalSelectionRule,
    HDF5CompositeCAEImportMode,
    HDF5CompositeCAEProjectionMode,
    IgnorableEntity,
    ImportedAnalysisPly,
    ImportedModelingGroup,
    ImportedModelingPly,
    ImportedPlyDrapingType,
    ImportedPlyOffsetType,
    ImportedPlyThicknessType,
    ImportedProductionPly,
    ImportedSolidModel,
    ImportedSolidModelExportSettings,
    InterfaceLayer,
    IntersectionType,
    Lamina,
    LayupMappingObject,
    LayupMappingRosetteSelectionMethod,
    LinkedSelectionRule,
    LookUpTable1D,
    LookUpTable1DColumn,
    LookUpTable3D,
    LookUpTable3DColumn,
    LookUpTable3DInterpolationAlgorithm,
    LookUpTableColumnValueType,
    Material,
    MeshImportType,
    Model,
    ModelingGroup,
    ModelingPly,
    NodalDataType,
    OffsetType,
    OrientedSelectionSet,
    ParallelSelectionRule,
    PhysicalDimension,
    PlyCutOffType,
    PlyGeometryExportFormat,
    PlyType,
    PrimaryPly,
    ProductionPly,
    ReinforcingBehavior,
    Rosette,
    RosetteSelectionMethod,
    RosetteType,
    SamplingPoint,
    SectionCut,
    SectionCutCDBExportType,
    SectionCutType,
    Sensor,
    SensorType,
    ShellMappingProperties,
    SnapToGeometry,
    SnapToGeometryOrientationType,
    SolidElementSet,
    SolidMappingProperties,
    SolidModel,
    SolidModelExportFormat,
    SolidModelExportSettings,
    SolidModelImportFormat,
    SolidModelOffsetDirectionType,
    SolidModelSkinExportFormat,
    SphericalSelectionRule,
    Stackup,
    Status,
    StressStateType,
    SubLaminate,
    SubShape,
    SymmetryType,
    TaperEdge,
    ThicknessFieldType,
    ThicknessType,
    TubeSelectionRule,
    UnitSystemType,
    VariableOffsetSelectionRule,
    VirtualGeometry,
    VirtualGeometryDimension,
)

__version__ = importlib.metadata.version(__name__.replace(".", "-"))


__all__ = [
    "__version__",
    "ACPInstance",
    "AnalysisPly",
    "ArrowType",
    "BaseElementMaterialHandling",
    "BooleanOperationType",
    "BooleanSelectionRule",
    "ButtJointSequence",
    "CADComponent",
    "CADGeometry",
    "ConnectLaunchConfig",
    "CoordinateTransformation",
    "CutOffGeometry",
    "CutOffGeometryOrientationType",
    "CutOffMaterialHandling",
    "CutOffRuleType",
    "CutOffSelectionRule",
    "CylindricalSelectionRule",
    "DirectLaunchConfig",
    "DockerComposeLaunchConfig",
    "dpf_integration_helpers",
    "DrapingMaterialModel",
    "DrapingType",
    "DropOffMaterialHandling",
    "DropOffSettings",
    "DropOffType",
    "EdgeSet",
    "EdgeSetType",
    "ElementalDataType",
    "ElementSet",
    "ElementTechnology",
    "ExportSettings",
    "extras",
    "ExtrusionGuide",
    "ExtrusionGuideType",
    "ExtrusionMethod",
    "ExtrusionType",
    "Fabric",
    "FabricWithAngle",
    "FeFormat",
    "FieldDefinition",
    "GeometricalRuleType",
    "GeometricalSelectionRule",
    "get_directions_plotter",
    "get_model_tree",
    "HDF5CompositeCAEImportMode",
    "HDF5CompositeCAEProjectionMode",
    "IgnorableEntity",
    "ImportedAnalysisPly",
    "ImportedModelingGroup",
    "ImportedModelingPly",
    "ImportedPlyDrapingType",
    "ImportedPlyOffsetType",
    "ImportedPlyThicknessType",
    "ImportedProductionPly",
    "ImportedSolidModel",
    "ImportedSolidModelExportSettings",
    "InterfaceLayer",
    "IntersectionType",
    "Lamina",
    "launch_acp",
    "LaunchMode",
    "LayupMappingObject",
    "LayupMappingRosetteSelectionMethod",
    "LinkedObjectHandling",
    "LinkedSelectionRule",
    "LookUpTable1D",
    "LookUpTable1DColumn",
    "LookUpTable3D",
    "LookUpTable3DColumn",
    "LookUpTable3DInterpolationAlgorithm",
    "LookUpTableColumnValueType",
    "material_property_sets",
    "Material",
    "mechanical_integration_helpers",
    "mesh_data",
    "MeshImportType",
    "Model",
    "ModelingGroup",
    "ModelingPly",
    "NodalDataType",
    "OffsetType",
    "OrientedSelectionSet",
    "ParallelSelectionRule",
    "PhysicalDimension",
    "PlyCutOffType",
    "PlyGeometryExportFormat",
    "PlyType",
    "PrimaryPly",
    "print_model",
    "ProductionPly",
    "recursive_copy",
    "ReinforcingBehavior",
    "Rosette",
    "RosetteSelectionMethod",
    "RosetteType",
    "SamplingPoint",
    "SectionCut",
    "SectionCutCDBExportType",
    "SectionCutType",
    "Sensor",
    "SensorType",
    "ShellMappingProperties",
    "SnapToGeometry",
    "SnapToGeometryOrientationType",
    "SolidElementSet",
    "SolidMappingProperties",
    "SolidModel",
    "SolidModelExportFormat",
    "SolidModelExportSettings",
    "SolidModelImportFormat",
    "SolidModelOffsetDirectionType",
    "SolidModelSkinExportFormat",
    "SphericalSelectionRule",
    "Stackup",
    "Status",
    "StressStateType",
    "SubLaminate",
    "SubShape",
    "SymmetryType",
    "TaperEdge",
    "ThicknessFieldType",
    "ThicknessType",
    "TubeSelectionRule",
    "UnitSystemType",
    "VariableOffsetSelectionRule",
    "VirtualGeometry",
    "VirtualGeometryDimension",
]



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import typing
from typing import Protocol

from typing_extensions import Self

from ansys.api.acp.v0.base_pb2 import ResourcePath

if typing.TYPE_CHECKING:  # pragma: no cover
    from ..base import ServerWrapper

__all__ = ["CreatableFromResourcePath", "tree_object_from_resource_path"]


class CreatableFromResourcePath(Protocol):
    """Interface for objects that can be created from a resource path."""

    @classmethod
    def _from_resource_path(
        cls, resource_path: ResourcePath, server_wrapper: ServerWrapper
    ) -> Self: ...


def tree_object_from_resource_path(
    resource_path: ResourcePath,
    server_wrapper: ServerWrapper,
    allowed_types: tuple[type[CreatableFromResourcePath], ...] | None = None,
) -> CreatableFromResourcePath | None:
    """Instantiate a tree object from its resource path.

    Parameters
    ----------
    resource_path :
        Resource path of the object.
    server_wrapper :
        Representation of the ACP server.
    allowed_types :
        Allowed types of the object. If None, all registered types are allowed.
    """
    #  Import here to avoid circular references. Cannot use the registry before
    #  all the object have been imported.
    from ..object_registry import object_registry

    # Resource path represents an object that is not set as an empty string
    # For instance fabric.material = None
    if resource_path.value == "":
        return None

    collection_name = resource_path.value.split("/")[::2][-1]
    resource_class: type[CreatableFromResourcePath] = object_registry[collection_name]
    if allowed_types is not None:
        if not issubclass(resource_class, allowed_types):
            raise TypeError(
                f"Resource path {resource_path.value} does not point to a valid "
                f"object type. Allowed types: {allowed_types}"
            )
    return resource_class._from_resource_path(resource_path, server_wrapper=server_wrapper)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator, MutableSequence
from functools import partial
import sys
from typing import Any, TypeVar, cast, overload

from grpc import Channel
import numpy as np
from typing_extensions import Self

from ansys.api.acp.v0.base_pb2 import ResourcePath

from .._object_cache import ObjectCacheMixin, constructor_with_cache
from ..base import TreeObject, TreeObjectBase
from .polymorphic_from_pb import tree_object_from_resource_path
from .property_helper import _exposed_grpc_property, _wrap_doc, grpc_data_getter, grpc_data_setter

ValueT = TypeVar("ValueT", bound=TreeObjectBase)


__all__ = ["LinkedObjectList", "define_linked_object_list", "define_polymorphic_linked_object_list"]


class LinkedObjectList(ObjectCacheMixin, MutableSequence[ValueT]):
    """List of linked tree objects."""

    @classmethod
    @constructor_with_cache(
        # NOTE greschd Feb'23:
        # We use the parent object's id() as part of the cache key since
        # the LinkedObjectList keeps its parent alive. This means the
        # id() does not change (and cannot be reused) as long as the
        # cache entry is alive.
        # This is somewhat incidental, but checked (indirectly) by the
        # object permanence tests. If we want to get rid of this, we
        # would instead need to find a way to handle the case where the
        # parent object is being stored.
        key_getter=lambda *args, parent_object, attribute_name, **kwargs: (
            id(parent_object),
            attribute_name,
        ),
        raise_on_invalid_key=True,
    )
    def _initialize_with_cache(
        cls: type[Self],
        *,
        parent_object: TreeObject,
        attribute_name: str,
        object_constructor: Callable[[ResourcePath, Channel], ValueT],
        allowed_types: tuple[type[ValueT], ...],
    ) -> Self:
        return cls(
            _parent_object=parent_object,
            _attribute_name=attribute_name,
            _object_constructor=object_constructor,
            _allowed_types=allowed_types,
        )

    @staticmethod
    def _cache_key_valid(key: Any) -> bool:
        try:
            (parent_object_id, attribute_name) = key
            if not attribute_name:
                return False
            if not isinstance(parent_object_id, int):
                return False
            return True
        except Exception:
            return False

    def __init__(
        self,
        *,
        _parent_object: TreeObject,
        _attribute_name: str,
        _object_constructor: Callable[[ResourcePath, Channel], ValueT],
        _allowed_types: tuple[Any, ...],
    ) -> None:
        getter = grpc_data_getter(_attribute_name, from_protobuf=list)
        setter = grpc_data_setter(_attribute_name, to_protobuf=lambda x: x)

        self._get_resourcepath_list = cast(
            Callable[[], list[ResourcePath]], lambda: getter(_parent_object)
        )

        def set_resourcepath_list(value: list[ResourcePath]) -> None:
            if not all([rp.value for rp in value]):
                # Check for empty resource paths
                raise RuntimeError("Cannot link to unstored objects.")
            setter(_parent_object, value)

        self._set_resourcepath_list = set_resourcepath_list
        self._object_constructor: Callable[[ResourcePath], ValueT] = (
            lambda resource_path: _object_constructor(resource_path, _parent_object._server_wrapper)
        )
        self._allowed_types = _allowed_types
        self._allowed_types_str = ", ".join([cls.__name__ for cls in _allowed_types])

    def __len__(self) -> int:
        return len(self._get_resourcepath_list())

    @overload
    def __getitem__(self, index: int) -> ValueT: ...

    @overload
    def __getitem__(self, index: slice) -> list[ValueT]: ...

    def __getitem__(self, index: int | slice) -> ValueT | list[ValueT]:
        resource_path = self._get_resourcepath_list()[index]
        if not isinstance(resource_path, list):
            assert isinstance(resource_path, ResourcePath)
            return self._object_constructor(resource_path)
        return [self._object_constructor(item) for item in resource_path]

    @overload
    def __setitem__(self, key: int, value: ValueT) -> None: ...

    @overload
    def __setitem__(self, key: slice, value: Iterable[ValueT]) -> None: ...

    def __setitem__(self, key: int | slice, value: ValueT | Iterable[ValueT]) -> None:
        resource_path_list = self._get_resourcepath_list()
        if isinstance(value, TreeObjectBase):
            self._check_type(value)
            if not isinstance(key, int):
                raise TypeError("Cannot assign to a slice with a single object.")
            resource_path_list[key] = value._resource_path
        else:
            if not isinstance(key, slice):
                raise TypeError("Cannot assign to a single index with an iterable.")
            for item in value:
                self._check_type(item)
            resource_path_list[key] = [item._resource_path for item in value]
        self._set_resourcepath_list(resource_path_list)

    def __delitem__(self, key: int | slice) -> None:
        resource_path_list = self._get_resourcepath_list()
        del resource_path_list[key]
        self._set_resourcepath_list(resource_path_list)

    def __iter__(self) -> Iterator[ValueT]:
        resource_path_list = self._get_resourcepath_list()
        yield from (self._object_constructor(item) for item in resource_path_list)

    def __reversed__(self) -> Iterator[ValueT]:
        resource_path_list = self._get_resourcepath_list()
        yield from (self._object_constructor(item) for item in reversed(resource_path_list))

    def __contains__(self, item: object) -> bool:
        return (
            hasattr(item, "_resource_path") and item._resource_path in self._get_resourcepath_list()
        )

    def append(self, object: ValueT) -> None:
        """Append an object to the list.

        Parameters
        ----------
        object:
            Object to append.
        """
        self._check_type(object)
        resource_path_list = self._get_resourcepath_list()
        resource_path_list.append(object._resource_path)
        self._set_resourcepath_list(resource_path_list)

    def count(self, value: ValueT) -> int:
        """Count the number of occurrences of an object in the list.

        Parameters
        ----------
        value:
            Object to count.
        """
        return self._get_resourcepath_list().count(value._resource_path)

    def index(self, value: ValueT, start: int = 0, stop: int = sys.maxsize) -> int:
        """Return the index of the first occurrence of an object in the list.

        Parameters
        ----------
        value:
            Object to find.
        """
        return self._get_resourcepath_list().index(value._resource_path, start, stop)

    def extend(self, iterable: Iterable[ValueT]) -> None:
        """Extend the list with an iterable of objects.

        Parameters
        ----------
        iterable:
            Iterable of objects to append.
        """
        resource_path_list = self._get_resourcepath_list()
        for it in iterable:
            self._check_type(it)
        resource_path_list.extend([it._resource_path for it in iterable])
        self._set_resourcepath_list(resource_path_list)

    def insert(self, index: int, object: ValueT) -> None:
        """Insert an object at a given index.

        Parameters
        ----------
        index:
            Index to insert at.
        object:
            Object to insert.
        """
        self._check_type(object)
        resource_path_list = self._get_resourcepath_list()
        resource_path_list.insert(index, object._resource_path)
        self._set_resourcepath_list(resource_path_list)

    def pop(self, index: int = -1) -> ValueT:
        """Remove and return an object from the list.

        Parameters
        ----------
        index:
            Index of the object to be removed.
        """
        resource_path_list = self._get_resourcepath_list()
        rp = resource_path_list.pop(index)
        self._set_resourcepath_list(resource_path_list)
        return self._object_constructor(rp)

    def remove(self, value: ValueT) -> None:
        """Remove the first occurrence of an object from the list.

        Parameters
        ----------
        value:
            Object to remove.
        """
        resource_path_list = self._get_resourcepath_list()
        resource_path_list.remove(value._resource_path)
        self._set_resourcepath_list(resource_path_list)

    def reverse(self) -> None:
        """Reverse the list in-place."""
        self._set_resourcepath_list(list(reversed(self._get_resourcepath_list())))

    def sort(
        self, *, key: Callable[[ValueT], Any] = lambda obj: obj.name, reverse: bool = False
    ) -> None:
        """Sort the list in-place.

        Parameters
        ----------
        key:
            Key function to sort by.
        reverse:
            Whether to sort in reverse order.
        """
        resource_path_list = self._get_resourcepath_list()
        evaluated_key_list = [key(self._object_constructor(rp)) for rp in resource_path_list]
        idx_list = np.argsort(evaluated_key_list)
        if reverse:
            idx_list = idx_list[::-1]
        resource_path_list = list(np.array(resource_path_list)[idx_list])
        self._set_resourcepath_list(resource_path_list)

    def __eq__(self, other: Any) -> Any:
        return list(self) == other

    def __repr__(self) -> str:
        return f"<LinkedObjectList([{', '.join(repr(val) for val in self)}])>"

    def _check_type(self, object: Any) -> None:
        if not isinstance(object, self._allowed_types):
            raise TypeError(
                f"List items must be of type {self._allowed_types_str}, not {type(object).__name__}."
            )


ParentT = TypeVar("ParentT", bound=TreeObject)
ChildT = TypeVar("ChildT", bound=TreeObjectBase)


def define_linked_object_list(
    attribute_name: str, object_class: type[ChildT], doc: str | None = None
) -> Any:
    """Define a list of linked tree objects."""

    def getter(self: ParentT) -> LinkedObjectList[ChildT]:
        return LinkedObjectList._initialize_with_cache(
            parent_object=self,
            attribute_name=attribute_name,
            object_constructor=object_class._from_resource_path,
            allowed_types=(object_class,),
        )

    def setter(self: ParentT, value: list[ChildT]) -> None:
        getter(self)[:] = value

    return _wrap_doc(_exposed_grpc_property(getter).setter(setter), doc=doc)


def define_polymorphic_linked_object_list(
    attribute_name: str,
    allowed_types: tuple[Any, ...] | None = None,
    allowed_types_getter: Callable[[], tuple[Any, ...]] | None = None,
) -> Any:
    """Define a list of linked tree objects with polymorphic types."""
    if allowed_types is None != allowed_types_getter is None:
        raise ValueError("Exactly one of allowed_types and allowed_types_getter must be provided.")

    def getter(self: ParentT) -> LinkedObjectList[Any]:
        nonlocal allowed_types
        if allowed_types_getter is not None:
            allowed_types = allowed_types_getter()

        assert allowed_types is not None
        return LinkedObjectList(
            _parent_object=self,
            _attribute_name=attribute_name,
            _object_constructor=partial(
                tree_object_from_resource_path, allowed_types=allowed_types
            ),
            _allowed_types=allowed_types,
        )

    def setter(self: ParentT, value: list[Any]) -> None:
        getter(self)[:] = value

    return _exposed_grpc_property(getter).setter(setter)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable
from functools import reduce

from google.protobuf.message import Message

from ansys.api.acp.v0 import extrusion_guide_pb2, extrusion_guide_pb2_grpc

from .._utils.array_conversions import to_1D_double_array, to_tuple_from_1D_array
from .._utils.property_protocols import ReadWriteProperty
from ._grpc_helpers.property_helper import (
    _PROTOBUF_T,
    _get_data_attribute,
    _set_data_attribute,
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from ._grpc_helpers.protocols import ObjectInfo
from .base import CreatableTreeObject, IdTreeObject
from .edge_set import EdgeSet
from .enums import (
    ExtrusionGuideType,
    extrusion_guide_type_from_pb,
    extrusion_guide_type_to_pb,
    status_type_from_pb,
)
from .object_registry import register
from .virtual_geometry import VirtualGeometry

# Workaround: these imports are needed to make sphinx_autodoc_typehints understand
# the inherited members of the Elemental- and NodalData classes.
import numpy as np  # noqa: F401 isort:skip

__all__ = [
    "ExtrusionGuide",
]


@mark_grpc_properties
@register
class ExtrusionGuide(CreatableTreeObject, IdTreeObject):
    """Instantiate an Extrusion Guide of a Solid Model.

    Parameters
    ----------
    name :
        The name of the Oriented Selection Set.
    active :
        Inactive extrusion guides are not used in the solid model extrusion.
    edge_set :
        Edge Set along which the Extrusion Guide acts.
    extrusion_guide_type :
        Type of the extrusion such as by direction or by geometry.
    cad_geometry  :
        CAD geometry along which the extrusion guide runs.
        Needed if the extrusion type is set to :attr:`ExtrusionGuideType.BY_GEOMETRY`.
    direction :
        Direction along which the extrusion guide runs. Need if
        the extrusion type is set to :attr:`ExtrusionGuideType.BY_DIRECTION`.
    radius :
        Controls the sphere of influence for mesh morphing.
    depth :
        Defines the bias of the mesh morphing.
    use_curvature_correction :
        Apply a curvature correction during the solid model extrusion which results in
        a smoother extruded surface. Under certain circumstances, deactivating
        curvature correction can lead to better extrusion results.

    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "extrusion_guides"
    _OBJECT_INFO_TYPE = extrusion_guide_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = extrusion_guide_pb2.CreateRequest
    _SUPPORTED_SINCE = "25.1"

    def __init__(
        self,
        *,
        name: str = "ExtrusionGuide",
        active: bool = True,
        edge_set: EdgeSet | None = None,
        extrusion_guide_type: ExtrusionGuideType = "by_direction",
        cad_geometry: VirtualGeometry | None = None,
        direction: tuple[float, float, float] = (0.0, 0.0, 1.0),
        radius: float = 0.0,
        depth: float = 0.0,
        use_curvature_correction: bool = False,
    ):
        super().__init__(name=name)
        self.active = active
        self.edge_set = edge_set
        self.extrusion_guide_type = ExtrusionGuideType(extrusion_guide_type)
        self.cad_geometry = cad_geometry
        self.direction = direction
        self.radius = radius
        self.depth = depth
        self.use_curvature_correction = use_curvature_correction

    def _create_stub(self) -> extrusion_guide_pb2_grpc.ObjectServiceStub:
        return extrusion_guide_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)

    active: ReadWriteProperty[bool, bool] = grpc_data_property("properties.active")

    edge_set = grpc_link_property("properties.edge_set", allowed_types=(EdgeSet,))

    extrusion_guide_type = grpc_data_property(
        "properties.extrusion_guide_type",
        from_protobuf=extrusion_guide_type_from_pb,
        to_protobuf=extrusion_guide_type_to_pb,
    )

    cad_geometry = grpc_link_property("properties.cad_geometry", allowed_types=(VirtualGeometry,))

    radius: ReadWriteProperty[float, float] = grpc_data_property("properties.radius")

    depth: ReadWriteProperty[float, float] = grpc_data_property("properties.depth")

    use_curvature_correction: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.use_curvature_correction"
    )

    # The extrusion guide type is not stored by the backend directly. Instead,
    # it is derived from the property direction. Therefore, setting and getting
    # the direction is blocked if the extrusion guide type is not `by_direction`.
    # See Feature 1122546 in ADO. Once resolved, this check becomes obsolete
    # but ensure backward compatibility.
    @staticmethod
    def _set_direction_attribute(pb_obj: ObjectInfo, name: str, value: _PROTOBUF_T) -> None:
        # name is "properties.direction"
        if (
            hasattr(pb_obj.properties, "extrusion_guide_type")
            and getattr(pb_obj.properties, "extrusion_guide_type")
            != extrusion_guide_pb2.BY_DIRECTION
        ):
            array = to_tuple_from_1D_array(value)
            if array and sum(array) != 0:
                raise RuntimeError(
                    "Cannot set direction if extrusion guide type is not 'by_direction'!"
                )
        _set_data_attribute(pb_obj, name, value)

    @staticmethod
    def _get_direction_attribute(pb_obj: Message, name: str, check_optional: bool) -> _PROTOBUF_T:
        # name is "properties.direction"
        name_parts = name.split(".")
        parent_obj = reduce(getattr, name_parts[:-1], pb_obj)
        if (
            hasattr(parent_obj, "extrusion_guide_type")
            and getattr(parent_obj, "extrusion_guide_type") != extrusion_guide_pb2.BY_DIRECTION
        ):
            raise RuntimeError(
                "Cannot access direction if the extrusion guide type is not 'by_direction'!"
            )
        return _get_data_attribute(pb_obj, name, check_optional)

    direction = grpc_data_property(
        "properties.direction",
        from_protobuf=to_tuple_from_1D_array,
        to_protobuf=to_1D_double_array,
        setter_func=_set_direction_attribute,
        getter_func=_get_direction_attribute,
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable

import numpy as np
import numpy.typing as npt

from ansys.api.acp.v0 import lookup_table_3d_column_pb2, lookup_table_3d_column_pb2_grpc

from ._grpc_helpers.property_helper import mark_grpc_properties
from .enums import LookUpTableColumnValueType, PhysicalDimension
from .lookup_table_column_base import LookUpTableColumnBase
from .object_registry import register

__all__ = ["LookUpTable3DColumn"]


@mark_grpc_properties
@register
class LookUpTable3DColumn(LookUpTableColumnBase):
    """Instantiate a Column of a 3D Look-Up Table.

    Parameters
    ----------
    value_type :
        Determines whether the column data is scalar (one entry per row) or
        directional (three entries per row).
        Note that the ``value_type`` can only be set when constructing the
        column, and is read-only afterwards.
    physical_dimension :
        Dimensionality (such as time, length, force, ...) of the column data.
    data :
        The column data. The shape of the data must match the ``value_type``
        and the length of the ``Location`` column of the parent look-up
        table.
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "lookup_table_3d_columns"
    _OBJECT_INFO_TYPE = lookup_table_3d_column_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = lookup_table_3d_column_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "LookUpTable3DColumn",
        value_type: LookUpTableColumnValueType = LookUpTableColumnValueType.SCALAR,
        physical_dimension: PhysicalDimension = PhysicalDimension.DIMENSIONLESS,
        data: npt.NDArray[np.float64] | None = None,
    ):
        super().__init__(
            name=name, value_type=value_type, physical_dimension=physical_dimension, data=data
        )

    def _create_stub(self) -> lookup_table_3d_column_pb2_grpc.ObjectServiceStub:
        return lookup_table_3d_column_pb2_grpc.ObjectServiceStub(self._channel)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Helpers for defining type annotations."""
import enum
import os
from typing import TYPE_CHECKING, Union

__all__ = ["PATH", "StrEnum"]

PATH = Union[str, os.PathLike[str]]

# For Python 3.10 and below, emulate the behavior of StrEnum by
# inheriting from str and enum.Enum.
# Note that this does *not* work on Python 3.11+, since the default
# Enum format method has changed and will not return the value of
# the enum member.
# When type checking, always use the Python 3.10 workaround, otherwise
# the StrEnum resolves as 'Any'.
if TYPE_CHECKING:  # pragma: no cover

    class StrEnum(str, enum.Enum):
        """String enum."""

else:
    try:
        from enum import StrEnum
    except ImportError:

        import enum

        class StrEnum(str, enum.Enum):
            """String enum."""

            pass



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Any, overload

from ..._grpc_helpers.property_helper import (
    _PROTOBUF_T,
    _exposed_grpc_property,
    grpc_data_property_read_only,
)

__all__ = ["variable_material_grpc_data_property", "constant_material_grpc_data_property"]

from ..._grpc_helpers.protocols import Editable, Readable


@overload
def variable_material_grpc_data_property(
    name: str, available_on_pb_type: None = None, unavailable_msg: None = None
) -> Any: ...


@overload
def variable_material_grpc_data_property(
    name: str,
    available_on_pb_type: type[_PROTOBUF_T],
    unavailable_msg: str,
) -> Any: ...


def variable_material_grpc_data_property(
    name: str,
    available_on_pb_type: type[_PROTOBUF_T] | None = None,
    unavailable_msg: str | None = None,
) -> Any:
    """Define a gRPC-backed property for a variable material property set."""

    def _from_protobuf(values: Any) -> tuple[Any]:
        if available_on_pb_type is not None:
            if not all(isinstance(val, available_on_pb_type.Data) for val in values):
                raise AttributeError(f"{name}: {unavailable_msg}")
        return tuple(getattr(val, name) for val in values)

    return grpc_data_property_read_only("values", from_protobuf=_from_protobuf)


def _constant_material_grpc_data_getter(
    name: str, available_on_pb_type: type[_PROTOBUF_T] | None, unavailable_msg: str | None
) -> Any:
    """Create a getter method which obtains the server object via the gRPC Get endpoint."""

    def inner(self: Readable) -> Any:
        self._get_if_stored()
        if available_on_pb_type is not None and not isinstance(
            self._pb_object, available_on_pb_type
        ):
            raise AttributeError(unavailable_msg)
        data_vals = self._pb_object.values
        if len(data_vals) != 1:
            raise RuntimeError(
                "The number of values is inconsistent with a constant material property."
            )
        return getattr(data_vals[0], name)

    return inner


def _constant_material_grpc_data_setter(
    name: str, available_on_pb_type: type[_PROTOBUF_T] | None, unavailable_msg: str | None
) -> Any:
    """Create a setter method which updates the server object via the gRPC Put endpoint."""

    def inner(self: Editable, value: Any) -> None:
        self._get_if_stored()
        if available_on_pb_type is not None and not isinstance(
            self._pb_object, available_on_pb_type
        ):
            raise AttributeError(f"{self.__class__.__name__}.{name}: {unavailable_msg}")
        data_vals = self._pb_object.values
        if len(data_vals) != 1:
            raise RuntimeError(
                "The number of values is inconsistent with a constant material property."
            )
        current_value = getattr(data_vals[0], name)
        try:
            needs_updating = current_value != value
        except TypeError:
            needs_updating = True
        if needs_updating:
            setattr(data_vals[0], name, value)
            self._put_if_stored()

    return inner


@overload
def constant_material_grpc_data_property(
    name: str, available_on_pb_type: None = None, unavailable_msg: None = None
) -> Any: ...


@overload
def constant_material_grpc_data_property(
    name: str,
    available_on_pb_type: type[_PROTOBUF_T],
    unavailable_msg: str,
) -> Any: ...


def constant_material_grpc_data_property(
    name: str,
    available_on_pb_type: type[_PROTOBUF_T] | None = None,
    unavailable_msg: str | None = None,
) -> Any:
    """Define a gRPC-backed property for a constant material property set."""
    return _exposed_grpc_property(
        _constant_material_grpc_data_getter(
            name=name, available_on_pb_type=available_on_pb_type, unavailable_msg=unavailable_msg
        )
    ).setter(
        _constant_material_grpc_data_setter(
            name=name, available_on_pb_type=available_on_pb_type, unavailable_msg=unavailable_msg
        )
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from enum import Enum

from ansys.api.acp.v0 import material_pb2

from ...._utils.property_protocols import ReadWriteProperty
from ..._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    mark_grpc_properties,
)
from ...base import TreeObject
from .base import _ConstantPropertySet, _VariablePropertySet
from .property_helper import (
    constant_material_grpc_data_property,
    variable_material_grpc_data_property,
)

__all__ = ["ConstantPuckConstants", "VariablePuckConstants", "PuckMaterialType"]


class PuckMaterialType(str, Enum):
    """Possible Puck material types."""

    IGNORED = "ignored"
    CARBON = "carbon"
    GLASS = "glass"
    MATERIAL_SPECIFIC = "material-specific"


class _PuckConstantsMixin:
    _DEFAULT_PB_PROPERTYSET_TYPE = material_pb2.PuckConstantsPropertySet


@mark_grpc_properties
class ConstantPuckConstants(_PuckConstantsMixin, _ConstantPropertySet):
    """Constant Puck constants material property set."""

    _GRPC_PROPERTIES = tuple()
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        p_21_pos: float | None = None,
        p_21_neg: float | None = None,
        p_22_pos: float | None = None,
        p_22_neg: float | None = None,
        s: float = 0.5,
        M: float = 0.5,
        interface_weakening_factor: float = 0.8,
        mat_type: str = PuckMaterialType.IGNORED,
        _parent_object: TreeObject | None = None,
        _attribute_path: str | None = None,
    ):
        super().__init__(_parent_object=_parent_object, _attribute_path=_attribute_path)
        if _parent_object is not None:
            return

        def val_or_default(val: float | None, default: float) -> float:
            if val is not None:
                return val
            return default

        self.mat_type = PuckMaterialType(mat_type)

        if mat_type in (PuckMaterialType.IGNORED, PuckMaterialType.MATERIAL_SPECIFIC):
            self.p_21_pos = val_or_default(p_21_pos, 0.325)
            self.p_21_neg = val_or_default(p_21_neg, 0.275)
            self.p_22_pos = val_or_default(p_22_pos, 0.225)
            self.p_22_neg = val_or_default(p_22_neg, 0.225)
        elif mat_type == PuckMaterialType.CARBON:
            self.p_21_pos = val_or_default(p_21_pos, 0.35)
            self.p_21_neg = val_or_default(p_21_neg, 0.3)
            self.p_22_pos = val_or_default(p_22_pos, 0.25)
            self.p_22_neg = val_or_default(p_22_neg, 0.25)
        elif mat_type == PuckMaterialType.GLASS:
            self.p_21_pos = val_or_default(p_21_pos, 0.3)
            self.p_21_neg = val_or_default(p_21_neg, 0.25)
            self.p_22_pos = val_or_default(p_22_pos, 0.2)
            self.p_22_neg = val_or_default(p_22_neg, 0.2)
        else:
            raise ValueError(f"Unknown 'mat_type': '{mat_type}'")

        self.s = s
        self.M = M
        self.interface_weakening_factor = interface_weakening_factor

    p_21_pos = constant_material_grpc_data_property("p_21_pos")
    p_21_neg = constant_material_grpc_data_property("p_21_neg")
    p_22_pos = constant_material_grpc_data_property("p_22_pos")
    p_22_neg = constant_material_grpc_data_property("p_22_neg")
    s = constant_material_grpc_data_property("s")
    M = constant_material_grpc_data_property("M")
    interface_weakening_factor = constant_material_grpc_data_property("interface_weakening_factor")
    mat_type: ReadWriteProperty[PuckMaterialType, PuckMaterialType] = grpc_data_property(
        "mat_type", from_protobuf=PuckMaterialType
    )


@mark_grpc_properties
class VariablePuckConstants(_PuckConstantsMixin, _VariablePropertySet):
    """Variable Puck constants material property set."""

    _GRPC_PROPERTIES = tuple()
    _SUPPORTED_SINCE = "24.2"

    p_21_pos = variable_material_grpc_data_property("p_21_pos")
    p_21_neg = variable_material_grpc_data_property("p_21_neg")
    p_22_pos = variable_material_grpc_data_property("p_22_pos")
    p_22_neg = variable_material_grpc_data_property("p_22_neg")
    s = variable_material_grpc_data_property("s")
    M = variable_material_grpc_data_property("M")
    interface_weakening_factor = variable_material_grpc_data_property("interface_weakening_factor")
    mat_type = grpc_data_property_read_only("mat_type", from_protobuf=PuckMaterialType)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Callable, Iterable, Sequence
from typing import TYPE_CHECKING, Any, Union, cast

from typing_extensions import Self

from ansys.api.acp.v0 import butt_joint_sequence_pb2, butt_joint_sequence_pb2_grpc

from .._utils.property_protocols import ReadWriteProperty
from ._grpc_helpers.edge_property_list import (
    GenericEdgePropertyType,
    define_add_method,
    define_edge_property_list,
)
from ._grpc_helpers.linked_object_list import define_polymorphic_linked_object_list
from ._grpc_helpers.polymorphic_from_pb import tree_object_from_resource_path
from ._grpc_helpers.property_helper import (
    _exposed_grpc_property,
    grpc_data_property,
    grpc_data_property_read_only,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .enums import status_type_from_pb
from .modeling_ply import ModelingPly
from .object_registry import register

if TYPE_CHECKING:  # pragma: no cover
    # Creates a circular import if imported at the top-level, since the ButtJointSequence
    # is a direct child of the ModelingGroup.
    from .modeling_group import ModelingGroup

__all__ = ["ButtJointSequence", "PrimaryPly"]


@mark_grpc_properties
class PrimaryPly(GenericEdgePropertyType):
    """Defines a primary ply of a butt joint sequence.

    Parameters
    ----------
    sequence :
        Modeling group or modeling ply defining the primary ply.
    level :
        Level of the primary ply. Plies with a higher level inherit the thickness
        from adjacent plies with a lower level.

    """

    _SUPPORTED_SINCE = "25.1"

    def __init__(self, sequence: ModelingGroup | ModelingPly, level: int = 1):
        self._callback_apply_changes: Callable[[], None] | None = None
        self.sequence = sequence
        self.level = level

    @_exposed_grpc_property
    def sequence(self) -> ModelingGroup | ModelingPly:
        """Linked sequence."""
        return self._sequence

    @sequence.setter
    def sequence(self, value: ModelingGroup | ModelingPly) -> None:
        from .modeling_group import ModelingGroup

        if not isinstance(value, (ModelingGroup, ModelingPly)):
            raise TypeError(f"Expected a ModelingGroup or ModelingPly, got {type(value)}")
        self._sequence = value
        if self._callback_apply_changes:
            self._callback_apply_changes()

    @_exposed_grpc_property
    def level(self) -> int:
        """Level of the primary ply.

        Plies with a higher level inherit the thickness from adjacent plies with a lower level.
        """
        return self._level

    @level.setter
    def level(self, value: int) -> None:
        self._level = value
        if self._callback_apply_changes:
            self._callback_apply_changes()

    def _set_callback_apply_changes(self, callback_apply_changes: Callable[[], None]) -> None:
        self._callback_apply_changes = callback_apply_changes

    @classmethod
    def _from_pb_object(
        cls,
        parent_object: CreatableTreeObject,
        message: butt_joint_sequence_pb2.PrimaryPly,
        apply_changes: Callable[[], None],
    ) -> Self:
        from .modeling_group import ModelingGroup  # imported here to avoid circular import

        new_obj = cls(
            sequence=cast(
                Union["ModelingGroup", ModelingPly],
                tree_object_from_resource_path(
                    message.sequence,
                    server_wrapper=parent_object._server_wrapper,
                    allowed_types=(ModelingGroup, ModelingPly),
                ),
            ),
            level=message.level,
        )
        new_obj._set_callback_apply_changes(apply_changes)
        return new_obj

    def _to_pb_object(self) -> butt_joint_sequence_pb2.PrimaryPly:
        return butt_joint_sequence_pb2.PrimaryPly(
            sequence=self.sequence._resource_path, level=self.level
        )

    def _check(self) -> bool:
        # Check for empty resource paths
        return bool(self.sequence._resource_path.value)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            return (
                self.sequence._resource_path == other.sequence._resource_path
                and self.level == other.level
            )

        return False

    def __repr__(self) -> str:
        return f"PrimaryPly(sequence={self.sequence.__repr__()}, level={self.level})"

    def clone(self) -> Self:
        """Create a new unstored PrimaryPly with the same properties."""
        return type(self)(sequence=self.sequence, level=self.level)


def _get_allowed_secondary_ply_types() -> tuple[type, ...]:
    from .modeling_group import ModelingGroup

    return (ModelingGroup, ModelingPly)


@mark_grpc_properties
@register
class ButtJointSequence(CreatableTreeObject, IdTreeObject):
    """Instantiate a ButtJointSequence.

    Parameters
    ----------
    name :
        Name of the butt joint sequence.
    primary_plies :
        Primary plies are the source of a butt joint and they pass the thickness to
        adjacent plies. Plies with a higher level inherit the thickness from those
        with a lower level.
    secondary_plies :
        Secondary plies are butt-joined to adjacent primary plies and they inherit
        the thickness.
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "butt_joint_sequences"
    _OBJECT_INFO_TYPE = butt_joint_sequence_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = butt_joint_sequence_pb2.CreateRequest
    _SUPPORTED_SINCE = "25.1"

    def __init__(
        self,
        *,
        name: str = "ButtJointSequence",
        active: bool = True,
        global_ply_nr: int = 0,
        primary_plies: Sequence[PrimaryPly] = (),
        secondary_plies: Sequence[ModelingGroup | ModelingPly] = (),
    ):
        super().__init__(name=name)
        self.active = active
        self.global_ply_nr = global_ply_nr
        self.primary_plies = primary_plies
        self.secondary_plies = secondary_plies

    def _create_stub(self) -> butt_joint_sequence_pb2_grpc.ObjectServiceStub:
        return butt_joint_sequence_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    active: ReadWriteProperty[bool, bool] = grpc_data_property("properties.active")
    global_ply_nr: ReadWriteProperty[int, int] = grpc_data_property("properties.global_ply_nr")

    primary_plies = define_edge_property_list("properties.primary_plies", PrimaryPly)
    add_primary_ply = define_add_method(
        PrimaryPly,
        attribute_name="primary_plies",
        func_name="add_primary_ply",
        parent_class_name="ButtJointSequence",
        module_name=__module__,
    )

    secondary_plies = define_polymorphic_linked_object_list(
        "properties.secondary_plies", allowed_types_getter=_get_allowed_secondary_ply_types
    )
